"""@package docstring
Iso2Mesh for Python - Primitive shape meshing functions

Copyright (c) 2024 Edward Xu <xu.ed at neu.edu>
              2024-2025 Qianqian Fang <q.fang at neu.edu>
"""

__all__ = [
    "getexeext",
    "fallbackexeext",
    "mwpath",
    "mcpath",
    "deletemeshfile",
    "rotatevec3d",
    "rotmat2vec",
    "varargin2struct",
    "jsonopt",
    "nargout",
]
##====================================================================================
## dependent libraries
##====================================================================================

import sys
import numpy as np
import os
import shutil
import platform
import re
import glob
import dis

ISO2MESH_BIN_VER = "1.9.8"

##====================================================================================
## implementations
##====================================================================================


def getexeext():
    osarch = platform.machine()
    ext = ".exe"
    if sys.platform == "linux":
        ext = ".mexa64"
    elif sys.platform.startswith("win"):
        ext = ".exe"
    elif sys.platform == "darwin" and osarch == "arm64":
        ext = ".mexmaca64"
    elif sys.platform == "darwin":
        ext = ".mexmaci64"
    else:
        print("Unable to find extension type")

    return ext


# _________________________________________________________________________________________________________


def fallbackexeext(exesuffix, exename):
    """
    Get the fallback external tool extension names for the current platform.

    Parameters:
        exesuffix: the output executable suffix from getexeext
        exename: the executable name

    Returns:
        exesuff: file extension for iso2mesh tool binaries
    """
    exesuff = exesuffix
    if exesuff == ".mexa64" and not os.path.isfile(
        mcpath(exename, exesuff)
    ):  # fall back to i386 linux
        exesuff = ".mexglx"
    if exesuff == ".mexmaci64" and not os.path.isfile(
        mcpath(exename, exesuff)
    ):  # fall back to i386 mac
        exesuff = ".mexmaci"
    if exesuff == ".mexmaci" and not os.path.isfile(
        mcpath(exename, exesuff)
    ):  # fall back to ppc mac
        exesuff = ".mexmac"
    if not os.path.isfile(mcpath(exename, exesuff)) and not os.path.isfile(
        os.path.join(mcpath(exename))
    ):  # fall back to OS native package
        exesuff = ""

    if not os.path.isfile(mcpath(exename, exesuff)) and not os.path.isfile(
        mcpath(exename)
    ):
        if shutil.which(exename):
            return exesuff
        raise FileNotFoundError(
            f"The following executable:\n\t{mcpath(exename)}{getexeext()}\n"
            "is missing. Please download it from "
            "https://github.com/fangq/iso2mesh/tree/master/bin/ "
            "and save it to the above path, then rerun the script.\n"
        )

    return exesuff


# _________________________________________________________________________________________________________


def mwpath(fname=""):
    """
    Get the full temporary file path by prepending the working directory
    and current session name.

    Parameters:
    fname : str, optional
        Input file name string (default is empty string).

    Returns:
    tempname : str
        Full file name located in the working directory.
    """

    # Retrieve the ISO2MESH_TEMP and ISO2MESH_SESSION environment variables
    p = os.getenv("ISO2MESH_TEMP")
    session = os.getenv("ISO2MESH_SESSION", "")

    # Get the current user's name for Linux/Unix/Mac/Windows
    username = os.getenv("USER") or os.getenv("UserName", "")
    if username:
        username = f"pyiso2mesh-{username}"

    tempname = ""

    if not p:
        tdir = os.path.abspath(
            os.path.join(os.sep, "tmp")
        )  # Use default temp directory
        if username:
            tdir = os.path.join(tdir, username)
            if not os.path.exists(tdir):
                os.makedirs(tdir)

        tempname = os.path.join(tdir, session, fname)
    else:
        tempname = os.path.join(p, session, fname)

    return tempname


# _________________________________________________________________________________________________________


def mcpath(fname, ext=None):
    """
    Get full executable path by prepending a command directory path.

    Parameters:
    fname : str
        Input file name string.
    ext : str, optional
        File extension.

    Returns:
    str
        Full file name located in the bin directory.
    """
    from pathlib import Path

    binname = ""

    # the bin folder under iso2mesh is searched first
    # tempname = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'bin', fname)
    tempname = os.path.join(os.path.expanduser("~"), "iso2mesh-tools")
    binfolder = Path(os.path.join(tempname, "iso2mesh-" + ISO2MESH_BIN_VER, "bin"))

    if os.path.isdir(tempname):
        binname = os.path.join(tempname, "iso2mesh-" + ISO2MESH_BIN_VER, "bin", fname)

        if ext:
            if os.path.isfile(binname + ext):
                binname = binname + ext
            else:
                binname = fname + ext

    else:
        import urllib.request
        import zipfile

        print("Iso2mesh meshing utilities do not exist locally, downloading now ...")
        os.makedirs(tempname)
        binurl = f"https://github.com/fangq/iso2mesh/archive/refs/tags/v{ISO2MESH_BIN_VER}.zip"
        filehandle, _ = urllib.request.urlretrieve(binurl)

        with zipfile.ZipFile(filehandle, "r") as zip_ref:
            for file in zip_ref.namelist():
                if file.startswith(f"iso2mesh-{ISO2MESH_BIN_VER}/bin/"):
                    zip_ref.extract(file, tempname)
                    extractfile = os.path.join(tempname, file)
                    print("Extracting " + extractfile)
                    if os.path.isfile(extractfile):
                        print("Setting permission " + extractfile)
                        os.chmod(extractfile, 0o755)
        if ext:
            binname = os.path.join(
                tempname, "iso2mesh-" + ISO2MESH_BIN_VER, "bin", fname + ext
            )
        else:
            binname = os.path.join(
                tempname, "iso2mesh-" + ISO2MESH_BIN_VER, "bin", fname
            )

    # on 64bit windows machine, try 'exename_x86-64.exe' first
    if (
        os.name == "nt"
        and "64" in os.environ["PROCESSOR_ARCHITECTURE"]
        and not re.search(r"_x86-64$", fname)
    ):
        w64bin = re.sub(r"(\.[eE][xX][eE])$", "_x86-64.exe", binname, count=1)
        if os.path.isfile(w64bin):
            binname = w64bin

    # if no such executable exist in iso2mesh/bin, find it in PATH env variable
    if "extractfile" not in locals() and ext and not os.path.isfile(binname):
        binname = fname

    return binname


# _________________________________________________________________________________________________________


def deletemeshfile(fname):
    """
    delete a given work mesh file under the working directory

    author: Qianqian Fang, <q.fang at neu.edu>

    input:
        fname: specified file name (without path)

    output:
        flag: not used
    """

    try:
        for f in glob.glob(fname):
            os.remove(f)
    except Exception as e:
        raise PermissionError(
            "You do not have permission to delete temporary files. If you are working in a multi-user "
            "environment, such as Unix/Linux and there are other users using iso2mesh, "
            "you may need to define ISO2MESH_SESSION='yourstring' to make your output "
            "files different from others; if you do not have permission to "
            f"{os.getcwd()} as the temporary directory, you have to define "
            "ISO2MESH_TEMP='/path/you/have/write/permission' in Python base workspace."
        ) from e


# _________________________________________________________________________________________________________


def rotatevec3d(pt, v1, u1=None, p0=None):
    """
    Rotate 3D points from one Cartesian coordinate system to another.

    Parameters:
    pt : numpy.ndarray
        3D points defined in a standard Cartesian system where a unitary
        z-vector is (0,0,1), 3 columns for x, y and z.
    v1 : numpy.ndarray
        The unitary z-vector for the target coordinate system.
    u1 : numpy.ndarray, optional
        The unitary z-vector for the source coordinate system, if ignored,
        u1=(0,0,1).
    p0 : numpy.ndarray, optional
        Offset of the new coordinate system, if ignored, p0=(0,0,0).

    Returns:
    newpt : numpy.ndarray
        The transformed 3D points.
    """

    if u1 is None:
        u1 = np.array([0, 0, 1])
    if p0 is None:
        p0 = np.array([0, 0, 0])

    u1 = u1 / np.linalg.norm(u1)
    v1 = v1 / np.linalg.norm(v1)

    R, s = rotmat2vec(u1.flatten(), v1.flatten())
    newpt = (R @ pt.T * s).T

    if p0 is not None:
        p0 = p0.flatten()
        newpt += np.tile(p0, (newpt.shape[0], 1))

    return newpt


# _________________________________________________________________________________________________________


def rotmat2vec(u, v):
    """
    [R,s]=rotmat2vec(u,v)

    the rotation matrix from vector u to v, satisfying R*u*s=v

    input:
      u: a 3D vector in the source coordinate system;
      v: a 3D vector in the target coordinate system;

    output:
      R: a rotation matrix to transform normalized u to normalized v
      s: a scaling factor, so that R*u*s=v
    """
    s = np.linalg.norm(v) / np.linalg.norm(u)
    u1 = u / np.linalg.norm(u)
    v1 = v / np.linalg.norm(v)

    k = np.cross(u1, v1)
    if not np.any(k):  # u and v are parallel
        R = np.eye(3)
        return R, s

    # Rodrigues's formula:
    costheta = np.dot(u1, v1)
    R = np.array([[0, -k[2], k[1]], [k[2], 0, -k[0]], [-k[1], k[0], 0]])
    R = costheta * np.eye(3) + R + np.outer(k, k) * (1 - costheta) / np.sum(k**2)

    return R, s


# _________________________________________________________________________________________________________


def varargin2struct(*args):
    opt = {}
    length = len(args)
    if length == 0:
        return opt

    i = 0
    while i < length:
        if isinstance(args[i], dict):
            opt = {**opt, **args[i]}  # Merging dictionaries
        elif isinstance(args[i], str) and i < length - 1:
            opt[args[i].lower()] = args[i + 1]
            i += 1
        else:
            raise ValueError(
                "input must be in the form of ...,'name',value,... pairs or structs"
            )
        i += 1

    return opt


# _________________________________________________________________________________________________________


def jsonopt(key, default, *args):
    val = default
    if len(args) <= 0:
        return val
    key0 = key.lower()
    opt = args[0]
    if isinstance(opt, dict):
        if key0 in opt:
            val = opt[key0]
        elif key in opt:
            val = opt[key]
    return val


def nargout():
    frame = sys._getframe(1)  # Get caller's frame
    code = frame.f_code
    lasti = frame.f_lasti

    # Look at the next instruction
    instruction = list(dis.get_instructions(code))[lasti // 2 + 1]

    if instruction.opname == "UNPACK_SEQUENCE":
        return instruction.arg
    else:
        return 1



"""@package docstring
Iso2Mesh for Python - Mesh data queries and manipulations

Copyright (c) 2024-2025 Qianqian Fang <q.fang at neu.edu>
"""

__all__ = [
    "finddisconnsurf",
    "surfedge",
    "volface",
    "extractloops",
    "meshconn",
    "nodevolume",
    "elemvolume",
    "neighborelem",
    "layersurf",
    "faceneighbors",
    "edgeneighbors",
    "maxsurf",
    "flatsegment",
    "mesheuler",
    "orderloopedge",
    "bbxflatsegment",
    "surfplane",
    "raytrace",
    "surfinterior",
    "surfpart",
    "surfseeds",
    "meshquality",
    "meshedge",
    "meshface",
    "surfacenorm",
    "nodesurfnorm",
    "uniqedges",
    "uniqfaces",
    "innersurf",
    "advancefront",
    "meshreorient",
    "meshcentroid",
    "elemfacecenter",
    "barydualmesh",
    "highordertet",
    "ismember_rows",
    "ray2surf",
]

##====================================================================================
## dependent libraries
##====================================================================================

import numpy as np
from itertools import combinations

##====================================================================================
## implementations
##====================================================================================


def finddisconnsurf(f):
    """
    Extract disconnected surfaces from a cluster of surfaces.

    Parameters:
    f : numpy.ndarray
        Faces defined by node indices for all surface triangles.

    Returns:
    facecell : list
        Separated disconnected surface node indices.
    """

    facecell = []  # Initialize output list
    subset = np.array([])  # Initialize an empty subset array

    # Loop until all faces are processed
    while f.size > 0:
        # Find the indices of faces connected to the first face
        idx = np.isin(f, f[0, :]).reshape(f.shape)
        ii = np.where(np.sum(idx, axis=1))[0]

        # Continue until all connected faces are processed
        while ii.size > 0:
            # Append connected faces to the subset
            subset = np.vstack((subset, f[ii, :])) if subset.size else f[ii, :]
            f = np.delete(f, ii, axis=0)  # Remove processed faces
            idx = np.isin(f, subset).reshape(f.shape)  # Update connection indices
            ii = np.where(np.sum(idx, axis=1))[0]  # Find next set of connected faces

        # If the subset is non-empty, append it to the output
        if subset.size > 0:
            facecell.append(subset)
            subset = np.array([])  # Reset subset

    return facecell


# _________________________________________________________________________________________________________


def surfedge(f, junction=None):
    """
    Find the edge of an open surface or surface of a volume.

    Parameters:
    f : numpy.ndarray
        Input surface facif f.size == 0:
        return np.array([]), None
    junction : int, optional
        If set to 1, allows finding junctions (edges with more than two connected triangles).

    Returns:
    openedge : numpy.ndarray
        List of edges of the specified surface.
    elemid : numpy.ndarray, optional
        Corresponding index of the tetrahedron or triangle with an open edge.
    """

    if f.size == 0:
        return np.array([]), None

    findjunc = 0

    if f.shape[1] == 3:
        edges = np.vstack(
            (f[:, [0, 1]], f[:, [1, 2]], f[:, [2, 0]])
        )  # create all the edges
    elif f.shape[1] == 4:
        edges = np.vstack(
            (f[:, [0, 1, 2]], f[:, [1, 0, 3]], f[:, [0, 2, 3]], f[:, [1, 3, 2]])
        )  # create all the edges
    else:
        raise ValueError("surfedge only supports 2D and 3D elements")

    edgesort = np.sort(edges, axis=1)
    _, ix, jx = np.unique(edgesort, axis=0, return_index=True, return_inverse=True)

    vec = np.bincount(jx, minlength=max(jx) + 1)
    if f.shape[1] == 3 and junction is not None:
        qx = np.where(vec > 2)[0]
    else:
        qx = np.where(vec == 1)[0]

    openedge = edges[ix[qx], :]

    elemid, _ = np.unravel_index(ix[qx], f.shape, order="F")

    return openedge, elemid + 1


# _________________________________________________________________________________________________________


def volface(t):
    """
    Find the surface patches of a volume.

    Parameters:
    t : numpy.ndarray
        Input volumetric element list (tetrahedrons), dimension (ne, 4).

    Returns:
    openface : numpy.ndarray
        List of faces of the specified volume.
    elemid : numpy.ndarray, optional
        The corresponding index of the tetrahedron with an open edge or triangle.
    """

    # Use surfedge function to find the boundary faces of the volume
    openface, elemid = surfedge(t)

    return openface, elemid


# _________________________________________________________________________________________________________


def extractloops(edges):
    """
    Extract individual loops or polyline segments from a collection of edges.

    Parameters:
    edges : numpy.ndarray
        A two-column matrix recording the starting/ending points of all edge segments.

    Returns:
    loops : list
        Output list of polyline or loop segments, with NaN separating each loop/segment.
    """

    loops = []

    # Remove degenerate edges (edges where the start and end points are the same)
    edges = edges[edges[:, 0] != edges[:, 1], :]

    if len(edges) == 0:
        return loops

    # Initialize the loop with the first edge
    loops.extend(edges[0, :])
    loophead = edges[0, 0]
    loopend = edges[0, 1]
    edges = np.delete(edges, 0, axis=0)

    while edges.size > 0:
        # Find the index of the edge connected to the current loop end
        idx = np.concatenate(
            [np.where(edges[:, 0] == loopend)[0], np.where(edges[:, 1] == loopend)[0]]
        )

        if len(idx) > 1:
            # If multiple connections found, select the first
            idx = idx[0]

        if not isinstance(idx, np.ndarray):
            idx = np.array(idx)

        if idx.size == 0:
            # If no connection found (open-line segment)
            idx_head = np.concatenate(
                [
                    np.where(edges[:, 0] == loophead)[0],
                    np.where(edges[:, 1] == loophead)[0],
                ]
            )
            if len(idx_head) == 0:
                # If both open ends are found, start a new loop
                loops.append(np.nan)
                loops.extend(edges[0, :])
                loophead = edges[0, 0]
                loopend = edges[0, 1]
                edges = np.delete(edges, 0, axis=0)
            else:
                # Flip and trace the other end of the loop
                loophead, loopend = loopend, loophead
                lp = np.flip(loops)
                seg = np.where(np.isnan(lp))[0]
                if len(seg) == 0:
                    loops = lp.tolist()
                else:
                    loops = (loops[: len(loops) - seg[0]] + lp[: seg[0]]).tolist()
            continue

        # Trace along a single line thread
        if idx.size == 1:
            ed = edges[idx, :].flatten()
            ed = ed[ed != loopend]
            newend = ed[0]
            if newend == loophead:
                # If a loop is completed
                loops.extend([loophead, np.nan])
                edges = np.delete(edges, idx, axis=0)
                if edges.size == 0:
                    break
                loops.extend(edges[0, :])
                loophead = edges[0, 0]
                loopend = edges[0, 1]
                edges = np.delete(edges, 0, axis=0)
                continue
            else:
                loops.append(newend)

            loopend = newend
            edges = np.delete(edges, idx, axis=0)

    return np.array(loops)


# _________________________________________________________________________________________________________


def meshconn(elem, nn):
    """
    Create a node neighbor list from a mesh.

    Parameters:
    elem : numpy.ndarray
        Element table of the mesh, where each row represents an element and its node indices.
    nn : int
        Total number of nodes in the mesh.

    Returns:
    conn : list
        A list of length `nn`, where each element is a list of all neighboring node IDs for each node.
    connnum : numpy.ndarray
        A vector of length `nn`, indicating the number of neighbors for each node.
    count : int
        Total number of neighbors across all nodes.
    """

    # Initialize conn as a list of empty lists
    conn = [[] for _ in range(nn)]
    dim = elem.shape
    # Loop through each element and populate the conn list
    for i in range(dim[0]):
        for j in range(dim[1]):
            conn[elem[i, j] - 1].extend(
                elem[i, :]
            )  # Adjust for 0-based indexing in Python

    count = 0
    connnum = np.zeros(nn, dtype=int)

    # Loop through each node to remove duplicates and self-references
    for i in range(nn):
        if len(conn[i]) == 0:
            continue
        # Remove duplicates and self-references
        neig = np.unique(conn[i])
        neig = neig[neig != i + 1]  # Remove self-reference, adjust for 0-based indexing
        conn[i] = neig.tolist()
        connnum[i] = len(conn[i])
        count += connnum[i]

    return conn, connnum, count


# _________________________________________________________________________________________________________


def nodevolume(node, elem, evol=None):
    """
    Calculate the volumes of the cells in the barycentric dual-mesh.
    This is different from Voronoi cells, which belong to the circumcentric dual mesh.

    Parameters:
    node : numpy.ndarray
        Node coordinates.
    elem : numpy.ndarray
        Element table of a mesh.
    evol : numpy.ndarray, optional
        Element volumes for each element (if not provided, it will be computed).

    Returns:
    nodevol : numpy.ndarray
        Volume values for all nodes.
    """

    # Determine if the mesh is 3D or 4D based on the number of nodes per element
    dim = 4 if elem.shape[1] == 4 else 3

    # If element volumes (evol) are not provided, calculate them
    if evol is None:
        evol = elemvolume(node, elem[:, :dim])

    elemnum = elem.shape[0]
    nodenum = node.shape[0]

    # Initialize node volume array
    nodevol = np.zeros(nodenum)

    # Loop through each element and accumulate the volumes
    for i in range(elemnum):
        nodevol[elem[i, :dim] - 1] += evol[i]

    # Divide by the dimensionality to get the final node volumes
    nodevol /= dim

    return nodevol


# _________________________________________________________________________________________________________


def elemvolume(node, elem, option=None):
    """
    vol = elemvolume(node, elem, option)

    Calculate the volume for a list of simplexes

    Parameters:
        node:   node coordinates (NumPy array)
        elem:   element table of a mesh (1-based indices)
        option: if option == 'signed', the volume is the raw determinant,
                otherwise, the result will be the absolute values

    Returns:
        vol: volume values for all elements
    """

    # Convert 1-based indices to 0-based for Python indexing
    v1 = node[elem[:, 0] - 1, :3]
    v2 = node[elem[:, 1] - 1, :3]
    v3 = node[elem[:, 2] - 1, :3]

    edge1 = v2 - v1
    edge2 = v3 - v1

    if elem.shape[1] == 3:
        # Triangle area in 2D or area in 3D projected onto a plane
        det12 = np.cross(edge1, edge2)
        det12 = np.sum(det12 * det12, axis=1)
        vol = 0.5 * np.sqrt(det12)
        return vol

    v4 = node[elem[:, 3] - 1, :3]
    edge3 = v4 - v1

    # Compute signed volume of tetrahedron
    vol = -np.einsum("ij,ij->i", edge1, np.cross(edge2, edge3, axis=1))

    if option == "signed":
        vol = vol / np.prod(np.arange(1, node.shape[1] + 1))
    else:
        vol = np.abs(vol) / np.prod(np.arange(1, node.shape[1] + 1))

    return vol


# _________________________________________________________________________________________________________


def neighborelem(elem, nn):
    """
    create node neighbor list from a mesh

    input:
       elem:  element table of a mesh
       nn  :  total node number of the mesh

    output:
       conn:  output, a list of length nn, conn[n]
              contains a list of all neighboring elem ID for node n
       connnum: list of length nn, denotes the neighbor number of each node
       count: total neighbor numbers
    """
    # Initialize conn as a list of empty lists
    conn = [[] for _ in range(nn)]
    dim = elem.shape

    elem = elem - 1
    # Loop through each element and populate the conn list
    for i in range(dim[0]):
        for j in range(dim[1]):
            conn[elem[i, j]].append(i + 1)  # Adjusting for 0-based index in Python

    # Loop through each node to sort neighbors and calculate total counts
    count = 0
    connnum = [0] * nn
    for i in range(nn):
        conn[i] = sorted(conn[i])
        connnum[i] = len(conn[i])
        count += connnum[i]

    return conn, connnum, count


# _________________________________________________________________________________________________________


def layersurf(elem, **kwargs):
    """
    face, labels = layersurf(elem, opt)
    or
    face, labels = layersurf(elem, option1=value1, option2=value2, ...)

    Process a multi-layered tetrahedral mesh to extract the layer surface meshes.

    Arguments:
    elem : an Nx5 integer array representing the tetrahedral mesh element list.
           The first 4 columns represent the tetrahedral element node indices,
           and the last column represents tissue labels.

    Optional kwargs:
    order : str, default '>=', options ['>=', '=', '<=']
        Determines how to process layers:
        '>=' (default): outmost layer has the lowest label count;
        '<=': innermost is lowest;
        '=': surface of each label is extracted individually.
    innermost : array-like, default [0]
        Labels treated as innermost regions, its boundary extracted using '=='.
        By default, label 0 is assumed to be the innermost (i.e., nothing enclosed inside).
    unique : bool, default False
        If True, removes duplicated triangles. If False, keeps all triangles.
    occurrence : str, default 'first', options ['first', 'last']
        If 'first', unique operator keeps the duplicated triangle with the lowest label number;
        otherwise, keeps the triangle with the highest label number.

    Returns:
    face : Nx4 array
        Extracted surface faces.
    labels : list
        Unique sorted labels in the mesh.
    """
    # Process input options
    opt = kwargs
    outsideislower = opt.get("order", ">=")
    dounique = opt.get("unique", False)
    innermost = opt.get("innermost", [0])
    occurrence = opt.get("occurrence", "first")

    labels = np.sort(np.unique(elem[:, 4]))
    face = []

    # Process each label
    for i in range(len(labels)):
        if outsideislower == ">=" and labels[i] not in innermost:
            newface = volface(elem[elem[:, 4] >= labels[i], :4])[0]
        elif outsideislower == "<=" and labels[i] not in innermost:
            newface = volface(elem[elem[:, 4] <= labels[i], :4])[0]
        else:
            newface = volface(elem[elem[:, 4] == labels[i], :4])[0]

        # Add label to faces
        newface = np.hstack((newface, np.full((newface.shape[0], 1), labels[i])))
        face.append(newface)

    face = np.vstack(face)

    # Remove duplicate triangles if unique option is enabled
    if dounique:
        face[:, :3] = np.sort(face[:, :3], axis=1)
        uniqface, idx = np.unique(face[:, :3], axis=0, return_index=True)
        face = np.hstack((uniqface, face[idx, -1].reshape(-1, 1)))

    return face, labels


# _________________________________________________________________________________________________________


def faceneighbors(t, opt=None):
    """
    facenb = faceneighbors(t, opt)

    Find the 4 face-neighboring elements of a tetrahedron.

    Arguments:
    t   : tetrahedron element list, 4 columns of integers.
    opt : if 'surface', return the boundary triangle list
          (same as face output from v2m).
          if 'rowmajor', return boundary triangles in row-major order.

    Output:
    facenb : If opt is 'surface', returns the list of boundary triangles.
             Otherwise, returns element neighbors for each element. Each
             row contains 4 numbers representing the element indices
             sharing triangular faces [1 2 3], [1 2 4], [1 3 4], and
             [2 3 4]. A 0 indicates no neighbor (i.e., boundary face).
    """
    # Generate faces from tetrahedral elements
    faces = np.vstack(
        (t[:, [0, 1, 2]], t[:, [0, 1, 3]], t[:, [0, 2, 3]], t[:, [1, 2, 3]])
    )
    faces = np.sort(faces, axis=1)

    # Find unique faces and their indices
    _, ix, jx = np.unique(faces, axis=0, return_index=True, return_inverse=True)

    vec = np.histogram(jx, bins=np.arange(max(jx) + 2))[0]
    qx = np.where(vec == 2)[0]

    nn = np.max(t)
    ne = t.shape[0]
    facenb = np.zeros_like(t)

    # Identify duplicate faces and their element pairings
    ujx, ii = np.unique(jx, return_index=True)
    jx2 = jx[::-1]
    _, ii2 = np.unique(jx2, return_index=True)
    ii2 = len(jx2) - 1 - ii2

    # List of element pairs that share a common face
    iddup = np.vstack((ii[qx], ii2[qx])).T
    faceid = np.ceil((iddup + 1) / ne).astype(int)
    eid = np.mod(iddup + 1, ne)
    eid[eid == 0] = ne

    # Handle special cases based on the second argument
    if opt is not None:
        for i in range(len(qx)):
            facenb[eid[i, 0] - 1, faceid[i, 0] - 1] = eid[i, 1]
            facenb[eid[i, 1] - 1, faceid[i, 1] - 1] = eid[i, 0]
        if opt == "surface":
            facenb = faces[np.where(facenb.T.flatten() == 0)[0], :]
        elif opt == "rowmajor":
            index = np.arange(len(faces)).reshape(4, -1).T.flatten()
            faces = faces[index, :]
            facenb = faces[np.where(facenb.flatten() == 0)[0], :]
        else:
            raise ValueError(f'Unsupported option "{opt}".')
    else:
        for i in range(len(qx)):
            facenb[eid[i, 0] - 1, faceid[i, 0] - 1] = eid[i, 1]
            facenb[eid[i, 1] - 1, faceid[i, 1] - 1] = eid[i, 0]

    return facenb


# _________________________________________________________________________________________________________


def edgeneighbors(t, opt=None):
    """
    edgenb = edgeneighbors(t, opt)

    Find neighboring triangular elements in a triangular surface.

    Arguments:
    t   : a triangular surface element list, 3 columns of integers.
    opt : (optional) If 'general', return edge neighbors for a general triangular surface.
          Each edge can be shared by more than 2 triangles. If ignored, assumes all
          triangles are shared by no more than 2 triangles.

    Output:
    edgenb : If opt is not supplied, edgenb is a size(t, 1) by 3 array, with each element
             being the triangle ID of the edge neighbor of that triangle. For each row,
             the neighbors are listed in the order of those sharing edges [1, 2], [2, 3],
             and [3, 1] between the triangle nodes.
             If opt = 'general', edgenb is a list of arrays, where each entry lists the edge neighbors.
    """
    # Generate the edges from the triangle elements
    edges = np.vstack([t[:, [0, 1]], t[:, [1, 2]], t[:, [2, 0]]])
    edges = np.sort(edges, axis=1)

    # Find unique edges and their indices
    _, ix, jx = np.unique(edges, axis=0, return_index=True, return_inverse=True)

    ne = t.shape[0]  # Number of triangles
    if opt == "general":
        edgenb = [
            np.unique(
                np.mod(
                    np.where(
                        (jx == jx[i]) | (jx == jx[i + ne]) | (jx == jx[i + 2 * ne])
                    )[0],
                    ne,
                )
            )
            for i in range(ne)
        ]
        return [np.setdiff1d(nb, [i]) for i, nb in enumerate(edgenb)]

    # Determine boundary neighbors
    vec = np.bincount(jx)
    qx = np.where(vec == 2)[
        0
    ]  # Get indices where edges are shared by exactly 2 triangles

    edgenb = np.zeros_like(t)

    ujx, first_idx = np.unique(jx, return_index=True)
    _, last_idx = np.unique(jx[::-1], return_index=True)
    last_idx = len(jx) - last_idx - 1

    # Find the element pairs that share an edge
    iddup = np.vstack([first_idx[qx], last_idx[qx]]).T
    faceid = (iddup // ne) + 1
    eid = iddup % ne
    eid += 1
    eid[eid == 0] = ne

    # Assign neighboring elements
    for i in range(len(qx)):
        edgenb[eid[i, 0] - 1, faceid[i, 0] - 1] = eid[i, 1]
        edgenb[eid[i, 1] - 1, faceid[i, 1] - 1] = eid[i, 0]

    # Handle boundary edges (where no neighbor exists)
    return edgenb


# _________________________________________________________________________________________________________


def maxsurf(facecell, node=None):
    """
    f, maxsize = maxsurf(facecell, node)

    Return the surface with the maximum number of elements or total area from a cell array of surfaces.

    Arguments:
    facecell : a list of arrays, each element representing a face array.
    node     : optional, node list. If given, the output is the surface with the largest surface area.

    Output:
    f        : the surface data (node indices) for the surface with the most elements (or largest area if node is given).
    maxsize  : if node is not provided, maxsize is the row number of f.
               If node is given, maxsize is the total area of f.
    """
    maxsize = -1
    maxid = -1

    # If node is provided, calculate area for each surface
    if node is not None:
        areas = np.zeros(len(facecell))
        for i in range(len(facecell)):
            areas[i] = np.sum(elemvolume(node[:, :3], facecell[i]))
        maxsize = np.max(areas)
        maxid = np.argmax(areas)
        f = facecell[maxid]
        return f, maxsize
    else:
        # Find the surface with the most elements
        for i in range(len(facecell)):
            if len(facecell[i]) > maxsize:
                maxsize = len(facecell[i])
                maxid = i

        f = []
        if maxid >= 0:
            f = facecell[maxid]

        return f, maxsize


def flatsegment(node, edge):
    """
    mask = flatsegment(node, edge)

    Decompose edge loops into flat segments along arbitrary planes of the bounding box.

    Arguments:
    node : Nx3 array of x, y, z coordinates for each node of the mesh.
    edge : vector separated by NaN, each segment is a closed polygon consisting of node IDs.

    Output:
    mask : list, each element is a closed polygon on the x/y/z plane.

    Author: Qianqian Fang
    Date: 2008/04/08

    Notes:
    This function is fragile: it cannot handle curves with many collinear nodes near corner points.
    """

    idx = edge
    nn = len(idx)
    val = np.zeros(nn)

    # Check for nearly flat tetrahedrons
    for i in range(nn):
        tet = np.mod(np.arange(i, i + 4), nn)
        tet[tet == 0] = nn
        # Calculate determinant to determine flatness
        val[i] = (
            abs(np.linalg.det(np.hstack((node[idx[tet], :], np.ones((4, 1)))))) > 1e-5
        )

    val = np.concatenate((val, val[:2]))
    mask = []
    oldend = 0
    count = 0

    # Decompose into flat segments
    for i in range(nn):
        if val[i] == 1 and val[i + 1] == 1 and val[i + 2] == 0:
            val[i + 2] = 2
            mask.append(idx[oldend : i + 3])
            count += 1
            oldend = i + 2
        else:
            mask.append(np.concatenate((idx[oldend:], mask[0])))
            break

    return mask


def mesheuler(face):
    """
    X, V, E, F, b, g, C = mesheuler(face)

    Compute Euler's characteristic of a mesh.

    Arguments:
    face : a closed surface mesh (Mx3 array where M is the number of faces and each row contains vertex indices)

    Output:
    X : Euler's characteristic (X = V - E + F - C)
    V : number of vertices
    E : number of edges
    F : number of triangles (if face is tetrahedral mesh, exterior surface)
    b : number of boundary loops (for surfaces)
    g : genus (holes)
    C : number of tetrahedra

    Author: Qianqian Fang
    This function is part of the iso2mesh toolbox (http://iso2mesh.sf.net)
    """

    # Number of vertices
    V = len(np.unique(face))

    # Number of unique edges
    E = uniqedges(face)[0].shape[0]

    b = 0  # open-boundary loops
    g = 0  # genus
    C = 0  # tet cells

    # Number of unique faces
    if face.shape[1] == 4:
        F = uniqfaces(face)[0].shape[0]
        C = face.shape[0]
    else:
        ed = surfedge(face)[0]
        loops = extractloops(ed)
        b = np.sum(np.isnan(loops))
        F = face.shape[0]

    # Euler's formula, X = V - E + F - C - 2*g
    X = V - E + F - C

    if face.shape[1] == 3:
        g = (X + b - 2) // 2

    return X, V, E, F, b, g, C


def orderloopedge(edge):
    """
    newedge = orderloopedge(edge)

    Order the node list of a simple loop based on the connection sequence.

    Arguments:
    edge : an Nx2 array where each row is an edge defined by two integers (start/end node index).

    Output:
    newedge : Nx2 array of reordered edge node list.

    Author: Qianqian Fang
    Date: 2008/05

    Notes:
    This function cannot process bifurcations.
    """

    ne = edge.shape[0]
    newedge = np.zeros_like(edge)
    newedge[0, :] = edge[0, :]

    for i in range(1, ne):
        row, col = np.where(edge[i:, :] == newedge[i - 1, 1])
        if len(row) == 1:
            newedge[i, :] = [newedge[i - 1, 1], edge[row[0] + i, 1 - col[0]]]
            edge[[i, row[0] + i], :] = edge[[row[0] + i, i], :]
        elif len(row) >= 2:
            raise ValueError("Bifurcation is found, exiting.")
        elif len(row) == 0:
            raise ValueError(f"Open curve at node {newedge[i - 1, 1]}")

    return newedge


def bbxflatsegment(node, edge):
    """
    mask = bbxflatsegment(node, edge)

    Decompose edge loops into flat segments along arbitrary planes of the bounding box.

    Arguments:
    node : Nx3 array of x, y, z coordinates for each node of the mesh.
    edge : vector separated by NaN, each segment is a closed polygon consisting of node IDs.

    Output:
    mask : list, each element is a closed polygon on the x/y/z plane.

    Author: Qianqian Fang
    Date: 2008/04/08

    Notes:
    This function is fragile: it cannot handle curves with many collinear nodes near corner points.
    """

    idx = edge
    nn = len(idx)
    val = np.zeros(nn)

    # Check for nearly flat tetrahedrons
    for i in range(nn):
        tet = np.mod(np.arange(i, i + 4), nn)
        tet[tet == 0] = nn
        # Calculate determinant to determine flatness
        val[i] = (
            abs(np.linalg.det(np.hstack((node[idx[tet], :], np.ones((4, 1)))))) > 1e-5
        )

    val = np.concatenate((val, val[:2]))
    mask = []
    oldend = 0
    count = 0

    # Decompose into flat segments
    for i in range(nn):
        if val[i] == 1 and val[i + 1] == 1 and val[i + 2] == 0:
            val[i + 2] = 2
            mask.append(idx[oldend : i + 3])
            count += 1
            oldend = i + 2
        else:
            mask.append(np.concatenate((idx[oldend:], mask[0])))
            break

    return mask


# _________________________________________________________________________________________________________


def surfplane(node, face):
    """
    plane = surfplane(node, face)

    Calculate plane equation coefficients for each face in a surface.

    Parameters:
    node : numpy array
        A list of node coordinates (nn x 3)
    face : numpy array
        A surface mesh triangle list (ne x 3)

    Returns:
    plane : numpy array
        A (ne x 4) array where each row has [a, b, c, d] to represent
        the plane equation as "a*x + b*y + c*z + d = 0"
    """

    # Compute vectors AB and AC from the triangle vertices
    AB = node[face[:, 1] - 1, :3] - node[face[:, 0] - 1, :3]
    AC = node[face[:, 2] - 1, :3] - node[face[:, 0] - 1, :3]

    # Compute normal vectors to the triangles using cross product
    N = np.cross(AB, AC)

    # Compute the plane's d coefficient by taking the dot product of normal vectors with vertex positions
    d = -np.sum(N * node[face[:, 0] - 1, :3], axis=1)

    # Return the plane coefficients [a, b, c, d]
    plane = np.column_stack((N, d))

    return plane


# _________________________________________________________________________________________________________


def raytrace(p0, v0, node, face):
    """
    t, u, v, idx = raytrace(p0, v0, node, face)

    Perform Havel-style ray tracing for a triangular surface.

    Parameters:
        p0: (3,) array, starting point of the ray
        v0: (3,) array, direction vector of the ray
        node: (nn, 3) array of node coordinates
        face: (ne, 3) array of triangle indices (1-based)

    Returns:
        t: signed distance to intersection (Inf if ray is parallel)
        u: barycentric coordinate 1
        v: barycentric coordinate 2
        idx: indices of intersected triangles (optional)
    """

    # Reshape p0 and v0 to row vectors
    p0 = np.asarray(p0).reshape(1, 3)
    v0 = np.asarray(v0).reshape(1, 3)

    # Convert 1-based indices in face to 0-based
    A = node[face[:, 0] - 1, :]
    B = node[face[:, 1] - 1, :]
    C = node[face[:, 2] - 1, :]

    AB = B - A
    AC = C - A

    # Normal vectors of triangles
    N = np.cross(AB, AC)
    d = -np.einsum("ij,ij->i", N, A)

    Rn2 = 1.0 / np.einsum("ij,ij->i", N, N)

    N1 = np.cross(AC, N) * Rn2[:, np.newaxis]
    d1 = -np.einsum("ij,ij->i", N1, A)

    N2 = np.cross(N, AB) * Rn2[:, np.newaxis]
    d2 = -np.einsum("ij,ij->i", N2, A)

    den = np.dot(N, v0.T).flatten()
    t = -(d + np.dot(N, p0.T).flatten())
    P = (np.outer(p0, den) + np.outer(v0, t)).T

    u = np.einsum("ij,ij->i", P, N1) + den * d1
    v = np.einsum("ij,ij->i", P, N2) + den * d2

    idx = den != 0
    den_inv = np.zeros_like(den)
    den_inv[idx] = 1.0 / den[idx]

    t = t * den_inv
    u = u * den_inv
    v = v * den_inv

    # For parallel rays, set t to Inf
    t[~idx] = np.inf

    # Compute intersection index if requested
    idx_out = None
    if u.shape[0] > 0 and v.shape[0] > 0 and t.shape[0] > 0:
        idx_out = np.where((u >= 0) & (v >= 0) & (u + v <= 1.0) & (~np.isinf(t)))[0]

    return t, u, v, idx_out


# _________________________________________________________________________________________________________


def surfinterior(node, face):
    """
    pt, p0, v0, t, idx = surfinterior(node, face)

    Identify a point that is enclosed by the (closed) surface.

    Arguments:
    node : a list of node coordinates (nn x 3)
    face : a surface mesh triangle list (ne x 3)

    Output:
    pt  : the interior point coordinates [x, y, z]
    p0  : ray origin used to determine the interior point
    v0  : the vector used to determine the interior point
    t   : ray-tracing intersection distances (with signs) from p0. The intersection coordinates
          can be expressed as p0 + t[i] * v0
    idx : index to the face elements that intersect with the ray, order matches that of t

    Author: Qianqian Fang
    This function is part of the iso2mesh toolbox (http://iso2mesh.sf.net)
    """

    pt, p0, v0, t, idx = [], [], [], None, []

    len_faces = face.shape[0]

    for i in range(len_faces):
        p0 = np.mean(
            node[face[i, :3] - 1, :], axis=0
        )  # Calculate the centroid of the triangle
        plane = surfplane(
            node, face[i, :].reshape(1, -1)
        )  # Plane equation for the current triangle
        v0 = plane[0][:3]  # Use the plane normal vector as the direction of the ray
        t, u, v, _ = raytrace(p0, v0, node, face[:, :3])  # Perform ray-tracing

        idx = np.where((u >= 0) & (v >= 0) & (u + v <= 1.0) & (~np.isinf(t)))[
            0
        ]  # Filter valid intersections

        # Sort and ensure ray intersections are valid
        ts, uidx = np.unique(np.sort(t[idx]), return_index=True)

        if len(ts) > 0 and len(ts) % 2 == 0:
            ts = ts.reshape((2, len(ts) // 2))
            tdiff = ts[1, :] - ts[0, :]
            maxi = np.argmax(tdiff)
            pt = (
                p0 + v0 * (ts[0, maxi] + ts[1, maxi]) * 0.5
            )  # Calculate the midpoint of the longest segment
            idx = idx[uidx]
            t = t[idx]
            break

    return pt, p0, v0, t, idx


def surfpart(f, loopedge):
    """
    elist = surfpart(f, loopedge)

    Partition a triangular surface using a closed loop defined by existing edges.

    Parameters:
    f : numpy array
        Surface face element list, dimension (n, 3) or (n, 4)
    loopedge : numpy array
        A 2-column array specifying a closed loop in counter-clockwise order.

    Returns:
    elist : numpy array
        List of triangles that is enclosed by the loop.
    """
    elist = []

    # Check if input is empty
    if f.size == 0 or loopedge.size == 0:
        return np.array(elist)

    # Handle triangular or quadrilateral elements
    if f.shape[1] == 3:
        # Create edges from triangles
        edges = np.vstack([f[:, [0, 1]], f[:, [1, 2]], f[:, [2, 0]]])
    elif f.shape[1] == 4:
        # Create edges from quadrilaterals
        edges = np.vstack([f[:, [0, 1]], f[:, [1, 2]], f[:, [2, 3]], f[:, [3, 0]]])
    else:
        raise ValueError("surfpart only supports triangular and quadrilateral elements")

    # Advance the front using the edges and loop
    elist, front = advancefront(edges, loopedge)

    # Continue advancing the front until no more elements can be added
    while front.size > 0:
        elist0, front0 = advancefront(edges, front)
        elist = np.unique(np.vstack([elist, elist0]), axis=0)
        front = front0

    return elist


def surfseeds(node, face):
    """
    seeds = surfseeds(node, face)

    Calculate a set of interior points, with each enclosed by a closed
    component of a surface.

    Parameters:
    node : numpy array
        A list of node coordinates (nn x 3).
    face : numpy array
        A surface mesh triangle list (ne x 3).

    Returns:
    seeds : numpy array
        Interior point coordinates for each closed surface component.
    """

    # Find disconnected surface components
    fc = finddisconnsurf(face[:, 0:3])
    len_fc = len(fc)

    # Initialize seed points array
    seeds = np.zeros((len_fc, 3))

    # For each disconnected component, calculate the interior point
    for i in range(len_fc):
        seeds[i, :] = surfinterior(node, fc[i])[0]

    return seeds


def meshquality(node, elem, maxnode=4):
    """
    quality = meshquality(node, elem, maxnode=4)

    Compute the Joe-Liu mesh quality measure of an N-D mesh (N <= 3).

    Parameters:
    node : numpy array
        Node coordinates of the mesh (nn x 3).
    elem : numpy array
        Element table of an N-D mesh (ne x (N+1)).
    maxnode : int, optional
        Maximum number of nodes per element (default is 4 for tetrahedral).

    Returns:
    quality : numpy array
        A vector of the same length as size(elem,1), with each element being
        the Joe-Liu mesh quality metric (0-1) of the corresponding element.
        A value close to 1 represents higher mesh quality (1 means equilateral tetrahedron);
        a value close to 0 means a nearly degenerated element.

    Reference:
    A. Liu, B. Joe, Relationship between tetrahedron shape measures,
    BIT 34 (2) (1994) 268-287.
    """

    if elem.shape[1] > maxnode:
        elem = elem[:, :maxnode]

    enum = elem.shape[0]

    # Compute element volume
    vol = elemvolume(node, elem)

    # Compute edge lengths
    edges = meshedge(elem)
    ed = node[edges[:, 0] - 1, :] - node[edges[:, 1] - 1, :]
    ed = np.sum(ed**2, axis=1)
    ed = np.sum(ed.reshape((enum, ed.size // enum), order="F"), axis=1)

    dim = elem.shape[1] - 1
    coeff = 10 / 9  # for tetrahedral elements

    if dim == 2:
        coeff = 1

    # Compute quality metric
    quality = (
        coeff
        * dim
        * 2 ** (2 * (1 - 1 / dim))
        * 3 ** ((dim - 1) / 2)
        * vol ** (2 / dim)
        / ed
    )

    # Normalize quality if max quality > 1
    maxquality = np.max(quality)
    if maxquality > 1:
        quality = quality / maxquality

    return quality


# _________________________________________________________________________________________________________


def meshedge(elem, opt=None):
    """
    edges = meshedge(elem, opt=None)

    Return all edges in a surface or volumetric mesh.

    Parameters:
    elem : numpy array
        Element table of a mesh (supports N-dimensional space elements).
    opt : dict, optional
        Optional input. If opt is provided as a dictionary, it can have the following field:
        - opt['nodeorder']: If 1, assumes the elem node indices are in CCW orientation;
                            if 0, uses combinations to order edges.

    Returns:
    edges : numpy array
        Edge list; each row represents an edge, specified by the starting and
        ending node indices. The total number of edges is size(elem,1) x comb(size(elem,2),2).
        All edges are ordered by looping through each element.
    """
    # Determine element dimension and the combination of node pairs for edges
    dim = elem.shape
    edgeid = np.array(list(combinations(range(dim[1]), 2)))
    len_edges = edgeid.shape[0]

    # Initialize edge list
    edges = np.zeros((dim[0] * len_edges, 2), dtype=elem.dtype)

    # Populate edges by looping through each element
    for i in range(len_edges):
        edges[i * dim[0] : (i + 1) * dim[0], :] = np.column_stack(
            (elem[:, edgeid[i, 0]], elem[:, edgeid[i, 1]])
        )

    return edges


# _________________________________________________________________________________________________________


def meshface(elem, opt=None):
    """
    faces = meshface(elem, opt=None)

    Return all faces in a surface or volumetric mesh.

    Parameters:
    elem : numpy array
        Element table of a mesh (supports N-dimensional space elements).
    opt : dict, optional
        Optional input. If provided, `opt` can contain the following field:
        - opt['nodeorder']: If 1, assumes the elem node indices are in CCW orientation;
                            otherwise, uses combinations to order faces.

    Returns:
    faces : numpy array
        Face list; each row represents a face, specified by node indices.
        The total number of faces is size(elem,1) x comb(size(elem,2),3).
        All faces are ordered by looping through each element.
    """
    dim = elem.shape
    faceid = np.array(list(combinations(range(dim[1]), 3)))
    len_faces = faceid.shape[0]

    # Initialize face list
    faces = np.zeros((dim[0] * len_faces, 3), dtype=elem.dtype)

    # Populate faces by looping through each element
    for i in range(len_faces):
        faces[i * dim[0] : (i + 1) * dim[0], :] = np.array(
            [elem[:, faceid[i, 0]], elem[:, faceid[i, 1]], elem[:, faceid[i, 2]]]
        ).T

    return faces


# _________________________________________________________________________________________________________


def surfacenorm(node, face, normalize=True):
    """
    snorm = surfacenorm(node, face, normalize=True)

    Compute the normal vectors for a triangular surface.

    Parameters:
    node : numpy array
        A list of node coordinates (nn x 3).
    face : numpy array
        A surface mesh triangle list (ne x 3).
    normalize : bool, optional
        If set to True, the normal vectors will be unitary (default is True).

    Returns:
    snorm : numpy array
        Output surface normal vector at each face.
    """

    # Compute the normal vectors using surfplane (function must be defined)
    snorm = surfplane(node, face)
    snorm = snorm[:, :3]

    # Normalize the normal vectors if requested
    if normalize:
        snorm = snorm / np.sqrt(np.sum(snorm**2, axis=1, keepdims=True))

    return snorm


# _________________________________________________________________________________________________________


def nodesurfnorm(node, elem):
    """
    nv = nodesurfnorm(node, elem)

    Calculate a nodal normal for each vertex on a surface mesh (the surface
    can only be triangular or cubic).

    Parameters:
    node : numpy array
        Node coordinates of the surface mesh (nn x 3).
    elem : numpy array
        Element list of the surface mesh (3 columns for triangular mesh,
        4 columns for cubic surface mesh).

    Returns:
    nv : numpy array
        Nodal normals calculated for each node (nn x 3).
    """

    nn = node.shape[0]  # Number of nodes
    ne = elem.shape[0]  # Number of elements
    nedim = elem.shape[1]  # Element dimension

    # Compute element normals
    ev = surfacenorm(node, elem)

    # Initialize nodal normals
    nv = np.zeros((nn, 3))

    # Sum element normals for each node
    for i in range(ne):
        nv[elem[i, :] - 1, :] += ev[i, :]

    # Normalize nodal normals
    nvnorm = np.linalg.norm(nv, axis=1)
    idx = np.where(nvnorm > 0)[0]

    if len(idx) < nn:
        print("Warning: Found interior nodes, their norms will be set to zeros.")
        nv[idx, :] = nv[idx, :] / nvnorm[idx][:, np.newaxis]
    else:
        nv = nv / nvnorm[:, np.newaxis]

    return nv


# _________________________________________________________________________________________________________


def uniqedges(elem):
    """
    edges, idx, edgemap = uniqedges(elem)

    Return the unique edge list from a surface or tetrahedral mesh.

    Parameters:
    elem : numpy array
        A list of elements, where each row is a list of nodes for an element.
        The input `elem` can have 2, 3, or 4 columns.

    Returns:
    edges : numpy array
        Unique edges in the mesh, denoted by pairs of node indices.
    idx : numpy array
        Indices of the unique edges in the raw edge list (returned by meshedge).
    edgemap : numpy array
        Index of the raw edges in the output list (for triangular meshes).
    """

    # Handle cases based on element size
    if elem.shape[1] == 2:
        edges = elem
    elif elem.shape[1] >= 3:
        edges = meshedge(elem)
    else:
        raise ValueError("Invalid input: element size not supported.")

    # Find unique edges and indices
    uedges, idx, jdx = np.unique(
        np.sort(edges, axis=1), axis=0, return_index=True, return_inverse=True
    )
    edges = edges[idx, :]

    # Compute edgemap if requested
    edgemap = np.reshape(
        jdx + 1,
        (-1, elem.shape[0]),
    )
    edgemap = edgemap.T

    return edges, idx + 1, edgemap


# _________________________________________________________________________________________________________


def uniqfaces(elem):
    """
    faces, idx, facemap = uniqfaces(elem)

    Return the unique face list from a surface or tetrahedral mesh.

    Parameters:
    elem : numpy array
        A list of elements, where each row contains node indices for an element.
        The input `elem` can have 2, 3, or 4 columns.

    Returns:
    faces : numpy array
        Unique faces in the mesh, denoted by triplets of node indices.
    idx : numpy array
        Indices of the unique faces in the raw face list (returned by meshface).
    facemap : numpy array
        Index of the raw faces in the output list (for triangular meshes).
    """

    # Determine faces based on element size
    if elem.shape[1] == 3:
        faces = elem
    elif elem.shape[1] >= 4:
        faces = meshface(elem)
    else:
        raise ValueError("Invalid input: element size not supported.")

    # Find unique faces and their indices
    ufaces, idx, jdx = np.unique(
        np.sort(faces, axis=1), axis=0, return_index=True, return_inverse=True
    )
    faces = faces[idx, :]

    # Compute facemap if requested
    facemap = np.reshape(
        jdx + 1,
        (elem.shape[0], np.array(list(combinations(range(elem.shape[1]), 3))).shape[0]),
        order="F",
    )

    return faces, idx + 1, facemap


def innersurf(node, face, outface=None):
    """
    Extract the interior triangles (shared by two enclosed compartments)
    of a complex surface.

    Parameters:
    node: Node coordinates
    face: Surface triangle list
    outface: (Optional) the exterior triangle list, if not provided,
             will be computed using outersurf().

    Returns:
    inface: The collection of interior triangles of the surface mesh
    """

    # If outface is not provided, compute it using outersurf
    if outface is None:
        outface = outersurf(node, face)

    # Check membership of sorted faces in sorted outface, row-wise
    tf, _ = ismember_rows(np.sort(face, axis=1), np.sort(outface, axis=1))

    # Select faces not part of the exterior (tf == 0)
    inface = face[tf == 0, :]

    return inface


def ismember_rows(array1, array2):
    # Ensure arrays are at least 2D and have same shape
    array1 = np.asarray(array1)
    array2 = np.asarray(array2)

    # Create structured view for row-wise comparison
    dtype = np.dtype((np.void, array1.dtype.itemsize * array1.shape[1]))
    a1_view = np.ascontiguousarray(array1).view(dtype)
    a2_view = np.ascontiguousarray(array2).view(dtype)

    isinside = np.isin(a1_view, a2_view)

    # Mapping: initialize with 0 (not found), +1 for 1-based MATLAB-like output
    map = np.zeros(array1.shape[0], dtype=int)
    for i in range(array1.shape[0]):
        matches = np.where((array2 == array1[i]).all(axis=1))[0]
        if matches.size > 0:
            map[i] = matches[0] + 1  # MATLAB-style index (1-based)

    return isinside, map


def advancefront(edges, loop, elen=3):
    """
    advance an edge-front on an oriented surface to the next separated by
    one-element width

    Author: Qianqian Fang
    Date: 2012/02/09

    Input:
    edges: edge list of an oriented surface mesh, must be in CCW order
    loop: a 2-column array, specifying a closed loop in CCW order
    elen: node number inside each element, if ignored, elen is set to 3

    Output:
    elist: list of triangles enclosed between the two edge-fronts
    nextfront: a new edge loop list representing the next edge-front
    """

    # Initialize output variables
    elist = []
    nextfront = []

    # Check if elen is provided, if not, set to 3
    if elen is None:
        elen = 3

    # Find edges that are part of the loop
    hasedge, loc = ismember(loop, edges)

    # If any edges in loop are not in the mesh, raise an error
    if not np.all(hasedge):
        raise ValueError("Loop edge is not defined in the mesh")

    # Calculate number of nodes in the mesh
    nodenum = len(edges) // elen

    # Find unique elements in the loop
    elist = np.unique((loc - 1) % nodenum) + 1

    # Get the corresponding edges for elist
    nextfront = edges[elist, :]

    # Loop through remaining elements
    for i in range(1, elen):
        nextfront = np.vstack([nextfront, edges[elist + nodenum * i, :]])

    # Remove reversed edge pairs
    nextfront = setdiff(nextfront, loop)
    flag, loc = ismember(nextfront, np.flip(nextfront, axis=1))

    id = np.where(flag)[0]
    if len(id) > 0:
        delmark = flag
        delmark[loc[loc > 0]] = 1
        nextfront = np.delete(nextfront, np.where(delmark), axis=0)

    # Reverse this loop as it is reversed relative to the input loop
    nextfront = nextfront[:, [1, 0]]

    return elist, nextfront


def ismember(A, B):
    """
    Check if rows of A are present in B.

    Returns a boolean array indicating membership and the corresponding indices.
    """
    return np.in1d(
        A.view([("", A.dtype)] * A.shape[1]), B.view([("", B.dtype)] * B.shape[1])
    ), np.where(np.in1d(A, B))


def setdiff(A, B):
    """
    Find the set difference between arrays A and B, row-wise.
    """
    dtype = np.dtype((np.void, A.dtype.itemsize * A.shape[1]))
    A_view = np.ascontiguousarray(A).view(dtype)
    B_view = np.ascontiguousarray(B).view(dtype)
    return A[~np.in1d(A_view, B_view)]


# _________________________________________________________________________________________________________


def meshreorient(node, elem):
    """
    Reorder nodes in a surface or tetrahedral mesh to ensure all
    elements are oriented consistently.

    Parameters:
        node: list of nodes
        elem: list of elements (each row are indices of nodes of each element)

    Returns:
        newelem: the element list with consistent ordering
        evol: the signed element volume before reorientation
        idx: indices of the elements that had negative volume
    """
    # Calculate the canonical volume of the element (can be a 2D or 3D)
    evol = elemvolume(node, elem, "signed")

    # Make sure all elements are positive in volume
    idx = np.where(evol < 0)[0]

    # Reorder the last two nodes for elements with negative volume
    elem[np.ix_(idx, [-2, -1])] = elem[np.ix_(idx, [-1, -2])]
    newelem = elem

    return newelem, evol, idx


# _________________________________________________________________________________________________________


def meshcentroid(v, f):
    #
    # centroid=meshcentroid(v,f)
    #
    # compute the centroids of a mesh defined by nodes and elements
    # (surface or tetrahedra) in R^n space
    #
    # input:
    #      v: surface node list, dimension (nn,3)
    #      f: surface face element list, dimension (be,3)
    #
    # output:
    #      centroid: centroid positions, one row for each element
    #
    if not isinstance(f, list):
        ec = v[f[:, :] - 1, :]
        centroid = np.squeeze(np.mean(ec, axis=1))
    else:
        length_f = len(f)
        centroid = np.zeros((length_f, v.shape[1]))
        try:
            for i in range(length_f):
                fc = f[i] - 1
                if fc:  # need to set centroid to NaN if fc is empty?
                    vlist = fc[0]
                    centroid[i, :] = np.mean(
                        v[vlist[~np.isnan(vlist)], :], axis=0
                    )  # Note to Ed check if this is functioning correctly
        except Exception as e:
            raise ValueError("malformed face cell array") from e
    return centroid


# _________________________________________________________________________________________________________


def elemfacecenter(node, elem):
    """
    Generate barycentric dual-mesh face center nodes and indices for each tetrahedral element.

    Args:
        node: List of node coordinates.
        elem: List of elements (each row contains the indices of nodes forming each tetrahedral element).

    Returns:
        newnode: Coordinates of new face-center nodes.
        newelem: Indices of the face-center nodes for each original tetrahedral element.
    """

    # Find unique faces from the elements (tetrahedral mesh)
    faces, idx, newelem = uniqfaces(elem[:, :4])

    # Extract the coordinates of the nodes forming these faces
    newnode = node[faces.T - 1, :3]
    newnode = np.mean(newnode, axis=0)

    return newnode, newelem


# _________________________________________________________________________________________________________


def barydualmesh(node, elem, flag=None):
    """
    Generate barycentric dual-mesh by connecting edge, face, and element centers.

    Parameters:
    node : numpy.ndarray
        List of input mesh nodes.
    elem : numpy.ndarray
        List of input mesh elements (each row contains indices of nodes for each element).
    flag : str, optional
        If 'cell', outputs `newelem` as cell arrays (each with 4 nodes).

    Returns:
    newnode : numpy.ndarray
        All new nodes in the barycentric dual-mesh (made of edge, face, and element centers).
    newelem : numpy.ndarray or list
        Indices of face nodes for each original tet element, optionally in cell array format.
    """

    # Compute edge-centers
    enodes, eidx = highordertet(node, elem)

    # Compute face-centers
    fnodes, fidx = elemfacecenter(node, elem)

    # Compute element centers
    c0 = meshcentroid(node, elem[:, : min(elem.shape[1], 4)])

    # Concatenate new nodes and their indices
    newnode = np.vstack((enodes, fnodes, c0))

    newidx = np.hstack(
        (
            eidx,
            fidx + enodes.shape[0],
            np.arange(1, elem.shape[0] + 1).reshape(-1, 1)
            + enodes.shape[0]
            + fnodes.shape[0],
        )
    )

    # Element connectivity for barycentric dual-mesh (using original indexing)
    newelem = (
        np.array(
            [
                [1, 8, 11, 7],
                [2, 7, 11, 9],
                [3, 9, 11, 8],
                [4, 7, 11, 10],
                [5, 8, 11, 10],
                [6, 9, 11, 10],
            ]
        ).T
        - 1
    )  # Adjust to 0-based indexing for Python

    newelem = newidx[:, newelem.flatten()]
    newelem = newelem.T.reshape(4, -1).T

    # If the 'cell' flag is set, return `newelem` as a list of lists (cells)
    if flag == "cell":
        newelem = [newelem[i, :].tolist() for i in range(newelem.shape[0])]

    return newnode, newelem


# _________________________________________________________________________________________________________


def highordertet(node, elem, order=2, opt=None):
    """
    Generate a higher-order tetrahedral mesh by refining a linear tetrahedral mesh.

    Args:
        node: Nodal coordinates of the linear tetrahedral mesh (n_nodes, 3).
        elem: Element connectivity (n_elements, 4).
        order: Desired order of the output mesh (default is 2 for quadratic mesh).
        opt: Optional dictionary to control mesh refinement options.

    Returns:
        newnode: Nodal coordinates of the higher-order tetrahedral mesh.
        newelem: Element connectivity of the higher-order tetrahedral mesh.
    """

    if order >= 3 or order <= 1:
        raise ValueError("currently this function only supports order=2")

    edges, idx, newelem = uniqedges(elem[:, : min(elem.shape[1], 4)])
    newnode = node[edges.T - 1, :3]  # adjust for 1-based index
    newnode = np.mean(newnode, axis=0)
    return newnode, newelem + 1


def internalpoint(v, aloop):
    """
    Empirical function to find an internal point of a planar polygon.

    Parameters:
        v : ndarray
            Nx3 array of x, y, z coordinates of each node of the mesh.
        aloop : array-like
            A vector of node indices (1-based, as in MATLAB), possibly separated by NaN.

    Returns:
        p : ndarray
            A single [x, y, z] internal point of the polygon loop.

    Raises:
        ValueError: If an internal point cannot be found.

    -- this function is part of iso2mesh toolbox (http://iso2mesh.sf.net)
    """
    from matplotlib.path import Path

    # Adjust aloop to 0-based index
    aloop = np.asarray(aloop, dtype=float)
    aloop = aloop[~np.isnan(aloop)].astype(int) - 1

    p = []
    nd = v[aloop, :]

    # Find if loop is flat along any axis
    boxfacet = np.where(np.sum(np.abs(np.diff(nd, axis=0)), axis=0) < 1e-2)[0]
    if len(boxfacet) > 0:
        bf = boxfacet[0]
        idx = [i for i in [0, 1, 2] if i != bf]

        p0 = (nd[0, :] + nd[1, :]) / 2
        pvec = complex(p0[idx[0]], p0[idx[1]])
        vec = nd[1, :] - nd[0, :]
        vec_mag = np.sqrt(np.sum(vec**2))
        vec = (
            complex(vec[idx[0]], vec[idx[1]])
            * np.exp(1j * np.pi / 2)
            * (1e-5)
            / vec_mag
        )
        testpt = np.array(
            [
                [np.real(pvec + vec), np.imag(pvec + vec)],
                [np.real(pvec - vec), np.imag(pvec - vec)],
            ]
        )

        path = Path(nd[:, idx])
        inside = path.contains_points(testpt)
        p2d = testpt[inside]
        if p2d.size > 0:
            p = np.zeros(3)
            p[[idx[0], idx[1]]] = p2d[0]
            p[bf] = nd[0, bf]

    if len(p) == 0 or len(p) != 3:
        raise ValueError("Fail to find an internal point of curve")

    return p


def ray2surf(node, elem, p0, v0, e0):
    """
    Determine the entry position and element for a ray to intersect a mesh

    Author: Qianqian Fang (q.fang <at> neu.edu)
    Python conversion: Preserves exact algorithm from MATLAB version

    Parameters:
    -----------
    node : ndarray
        The mesh coordinate list
    elem : ndarray
        The tetrahedral mesh element list, 4 columns
    p0 : ndarray
        Origin of the ray
    v0 : ndarray
        Direction vector of the ray
    e0 : str or float
        Search direction: '>' forward search, '<' backward search, '-' bidirectional

    Returns:
    --------: np.ndarray
    p : ndarray
        The intersection position
    e0 : int or float
        If found, the index of the intersecting element ID

    Notes:
    ------
    This file is part of Mesh-based Monte Carlo (MMC)
    License: GPLv3, see http://mcx.sf.net/mmc/ for details
    """

    p = p0.copy()

    if elem.shape[1] == 3:
        face = elem.copy()
    else:
        face = volface(elem)

    t, u, v, idx = raytrace(p0, v0, node, face)

    if len(idx) == 0:  # isempty(idx) in MATLAB
        raise RuntimeError("ray does not intersect with the mesh")
    else:
        t = t[idx]
        if e0 == ">":
            # idx1 = find(t>=0);
            idx1 = np.where(t >= 1e-10)[0]
        elif e0 == "<":
            idx1 = np.where(t <= 0)[0]
        elif np.isnan(e0) or e0 == "-":
            idx1 = np.arange(len(t))
        else:
            raise ValueError("ray direction specifier is not recognized")

        if len(idx1) == 0:  # isempty(idx1) in MATLAB
            raise RuntimeError("no intersection is found along the ray direction")

        t0 = np.abs(t[idx1])
        loc = np.argmin(t0)
        tmin = t0[loc]
        faceidx = idx[idx1[loc]]

        # Update source position
        p = p0 + t[idx1[loc]] * v0

        if elem.shape[1] == 3:
            e0 = faceidx
        else:
            # Convert faceidx to 0-based index when using as array index
            felem = np.sort(face[faceidx, :])
            f = elem.copy()

            # Create face combinations - subtract 1 for 0-based indexing when accessing elem
            f = np.vstack(
                [
                    elem[
                        :, [0, 1, 2]
                    ],  # elem[:, [1, 2, 3]] - 1 (MATLAB 1-based to Python 0-based)
                    elem[:, [1, 0, 3]],  # elem[:, [2, 1, 4]] - 1
                    elem[:, [0, 2, 3]],  # elem[:, [1, 3, 4]] - 1
                    elem[:, [1, 3, 2]],  # elem[:, [2, 4, 3]] - 1
                ]
            )

            # Sort each row for comparison
            f_sorted = np.sort(f, axis=1)

            # Find matching face using ismember equivalent
            tf, loc = ismember_rows(felem.reshape(1, -1), f_sorted)

            if tf[0]:
                loc = loc[0] % elem.shape[0]
                if loc == 0:
                    loc = elem.shape[0]
                e0 = loc
            else:
                # If no match found, return original faceidx
                e0 = faceidx

    return p, e0



"""@package docstring
Iso2Mesh for Python - Mesh-to-volume mesh rasterization

Copyright (c) 2024-2025 Qianqian Fang <q.fang at neu.edu>
"""

__all__ = ["m2v", "mesh2vol", "mesh2mask", "barycentricgrid"]

##====================================================================================
## dependent libraries
##====================================================================================

import numpy as np
import matplotlib.pyplot as plt
from iso2mesh.modify import qmeshcut

##====================================================================================
## implementations
##====================================================================================


def m2v(*args, **kwargs):
    """
    Shortcut for mesh2vol, rasterizing a tetrahedral mesh to a volume.

    Parameters:
    Same as mesh2vol function.

    Returns:
    Volumetric representation of the mesh.
    """
    return mesh2vol(*args, **kwargs)


def mesh2vol(node, elem, xi, yi=None, zi=None, **kwargs):
    """
    mesh2vol(node, elem, xi, yi=None, zi=None)

    Fast rasterization of a 3D tetrahedral mesh into a volumetric label image.

    Parameters:
        node : ndarray
            Node coordinates (N x 3) or (N x 4, with values in 4th column)
        elem : ndarray
            Tetrahedral elements (M x 4 or M x >4)
        xi, yi, zi : array-like or scalar
            Grid definitions. Supports:
              - scalar: voxel resolution
              - [Nx, Ny, Nz]: volume size
              - xi, yi, zi: actual grid vectors

    Returns:
        mask : 3D ndarray
            Voxelized volume with element labels
        weight : 4 x Nx x Ny x Nz array (if requested or values present)

    Author:
        Qianqian Fang <q.fang at neu.edu>
    """

    node = np.array(node, dtype=np.float64)
    elem = np.array(elem, dtype=np.int32)

    nodeval = None
    if node.shape[1] == 4:
        nodeval = node[:, 3].copy()
        node = node[:, :3]

    if yi is None and zi is None:
        if isinstance(xi, (int, float)):
            mn = np.min(node, axis=0)
            mx = np.max(node, axis=0)
            df = (mx - mn) / xi
        elif isinstance(xi, (list, tuple, np.ndarray)) and len(xi) == 3:
            mn = np.min(node, axis=0)
            mx = np.max(node, axis=0)
            df = (mx - mn) / np.array(xi)
        else:
            raise ValueError(
                "xi must be scalar or 3-element vector if yi and zi are not provided"
            )
        xi = np.arange(mn[0], mx[0] + df[0], df[0])
        yi = np.arange(mn[1], mx[1] + df[1], df[1])
        zi = np.arange(mn[2], mx[2] + df[2], df[2])
    else:
        xi = np.array(xi)
        yi = np.array(yi)
        zi = np.array(zi)
        df = [np.min(np.diff(xi)), np.min(np.diff(yi)), np.min(np.diff(zi))]

    if node.shape[1] != 3 or elem.shape[1] < 4:
        raise ValueError("node must have 3 columns; elem must have 4 or more columns")

    nx, ny, nz = len(xi), len(yi), len(zi)
    mask = np.zeros((nx, ny, nz))
    weight = np.zeros((4, nx, ny, nz)) if nodeval is not None else None

    for i, zval in enumerate(zi[:-1]):
        if nodeval is not None:
            cutpos, cutvalue, facedata, elemid, _ = qmeshcut(
                elem, node, nodeval, f"z={zval}"
            )
        else:
            cutpos, cutvalue, facedata, elemid, _ = qmeshcut(
                elem, node, node[:, 0], f"z={zval}"
            )
        if cutpos is None or len(cutpos) == 0:
            continue

        if weight is not None:
            maskz, weightz = mesh2mask(cutpos, facedata, xi, yi, **kwargs)
            weight[:, :, :, i] = weightz
        else:
            maskz = mesh2mask(cutpos, facedata, xi, yi, **kwargs)[0]

        idx = ~np.isnan(maskz)
        if nodeval is not None:
            eid = facedata[maskz[idx].astype(int) - 1]  # 1-based to 0-based
            maskz_flat = (
                cutvalue[eid[:, 0]] * weightz[0, idx]
                + cutvalue[eid[:, 1]] * weightz[1, idx]
                + cutvalue[eid[:, 2]] * weightz[2, idx]
                + cutvalue[eid[:, 3]] * weightz[3, idx]
            )
            maskz[idx] = maskz_flat
        else:
            maskz[idx] = elemid[(maskz[idx] - 1).astype(int)]  # adjust 1-based index

        mask[:, :, i] = maskz

    return mask, weight


def mesh2mask(node, face, xi, yi=None, hf=None, **kwargs):
    """
    Fast rasterization of a 2D mesh to an image with triangle index labels.

    Parameters:
    node: Node coordinates (N by 2 or N by 3 array)
    face: Triangle surface (N by 3 or N by 4 array)
    xi: Grid or number of divisions along x-axis
    yi: (Optional) Grid along y-axis
    hf: (Optional) Handle to a figure for faster rendering

    Returns:
    mask: 2D image where pixel values correspond to the triangle index
    weight: (Optional) Barycentric weights for each triangle
    """
    from matplotlib.collections import PatchCollection
    from matplotlib.patches import Polygon
    import matplotlib.cm as cm

    # Determine grid size from inputs
    if isinstance(xi, (int, float)) and yi is None:
        mn = np.min(node, axis=0)
        mx = np.max(node, axis=0)
        df = (mx[:2] - mn[:2]) / xi
    elif len(xi) == 2 and yi is None:
        mn = np.min(node, axis=0)
        mx = np.max(node, axis=0)
        df = (mx[:2] - mn[:2]) / xi
    elif yi is not None:
        mx = [np.max(xi), np.max(yi)]
        mn = [np.min(xi), np.min(yi)]
        df = [np.min(np.diff(xi)), np.min(np.diff(yi))]
    else:
        raise ValueError("At least xi input is required")

    # Error checking for input sizes
    if node.shape[1] <= 1 or face.shape[1] <= 2:
        raise ValueError(
            "node must have 2 or 3 columns; face must have at least 3 columns"
        )

    fig = (
        plt.figure(figsize=(xi.size * 0.01, yi.size * 0.01), dpi=100)
        if hf is None
        else hf
    )
    ax = fig.add_subplot(111)
    ax.set_position([0, 0, 1, 1])
    ax.set_xlim(mn[0], mx[0])
    ax.set_ylim(mn[1], mx[1])
    ax.set_axis_off()

    colors = cm.jet(np.linspace(0, 1, len(face)))

    patches = []
    for i, f in enumerate(face[:, :3]):
        polygon = Polygon(node[f - 1, :2], closed=True, zorder=1)
        patches.append(polygon)

    collection = PatchCollection(
        patches,
        facecolors=colors,
        linewidths=0,
        edgecolor="face",
        antialiased=(not kwargs.get("edge", True)),
    )
    ax.add_collection(collection)

    plt.draw()
    fig.canvas.draw()
    img = np.array(fig.canvas.renderer.buffer_rgba())

    mask = np.zeros(img.shape[:2], dtype=np.int32) * np.nan
    color_vals = np.floor(colors[:, :3] * 255 + 0.5).astype(np.uint8)

    for idx, cval in enumerate(color_vals):
        match = np.all(img[:, :, :3] == cval, axis=-1)
        mask[match] = idx + 1

    mask = mask[: len(yi), : len(xi)].T
    weight = barycentricgrid(node, face, xi, yi, mask)

    if hf is None:
        plt.close(fig)
    return mask, weight


def barycentricgrid(node, face, xi, yi, mask):
    """
    Compute barycentric weights for a 2D triangle mesh over a pixel grid.

    Parameters:
        node : ndarray (N, 2 or 3)
            Node coordinates.
        face : ndarray (M, 3)
            Triangle face indices (1-based).
        xi, yi : 1D arrays
            Grid coordinate vectors.
        mask : 2D ndarray
            Label image where each pixel contains the triangle index (1-based), NaN if outside.

    Returns:
        weight : ndarray (3, H, W)
            Barycentric coordinate weights for each pixel inside a triangle.
    """
    xx, yy = np.meshgrid(xi, yi, indexing="ij")  # shape: (H, W)
    mask = mask.astype(float)
    valid_idx = ~np.isnan(mask)

    # 1-based to 0-based index
    eid = mask[valid_idx].astype(int) - 1

    # triangle vertices (all triangles)
    t1 = node[face[:, 0] - 1]
    t2 = node[face[:, 1] - 1]
    t3 = node[face[:, 2] - 1]

    # denominator (twice the area of each triangle)
    tt = (t2[:, 1] - t3[:, 1]) * (t1[:, 0] - t3[:, 0]) + (t3[:, 0] - t2[:, 0]) * (
        t1[:, 1] - t3[:, 1]
    )

    # numerator for w1 and w2 (barycentric weights)
    w1 = (t2[eid, 1] - t3[eid, 1]) * (xx[valid_idx] - t3[eid, 0]) + (
        t3[eid, 0] - t2[eid, 0]
    ) * (yy[valid_idx] - t3[eid, 1])
    w2 = (t3[eid, 1] - t1[eid, 1]) * (xx[valid_idx] - t3[eid, 0]) + (
        t1[eid, 0] - t3[eid, 0]
    ) * (yy[valid_idx] - t3[eid, 1])

    w1 = w1 / tt[eid]
    w2 = w2 / tt[eid]
    w3 = 1 - w1 - w2

    # Assemble the weight volume
    weight = np.zeros((3, *mask.shape), dtype=np.float32)
    ww = np.zeros_like(mask, dtype=np.float32)

    ww[valid_idx] = w1
    weight[0, :, :] = ww
    ww[valid_idx] = w2
    weight[1, :, :] = ww
    ww[valid_idx] = w3
    weight[2, :, :] = ww

    return weight



"""@package docstring
Iso2Mesh for Python - Mesh data queries and manipulations

Copyright (c) 2024-2025 Qianqian Fang <q.fang at neu.edu>
"""
__all__ = [
    "v2m",
    "v2s",
    "s2m",
    "s2v",
    "vol2mesh",
    "vol2surf",
    "surf2mesh",
    "surf2volz",
    "surf2vol",
    "binsurface",
    "cgalv2m",
    "cgals2m",
    "getintersecttri",
    "vol2restrictedtri",
    "fillsurf",
    "outersurf",
    "surfvolume",
    "insurface",
    "remeshsurf",
    "meshrefine",
]

##====================================================================================
## dependent libraries
##====================================================================================

import numpy as np
import os
import re
import sys
import subprocess
from iso2mesh.trait import (
    surfinterior,
    surfseeds,
    surfedge,
    elemvolume,
    meshreorient,
    finddisconnsurf,
    maxsurf,
    volface,
    surfseeds,
    maxsurf,
    ismember_rows,
)
from iso2mesh.utils import *
from iso2mesh.io import (
    saveoff,
    readoff,
    saveinr,
    readtetgen,
    savesurfpoly,
    readmedit,
    savetetgennode,
    savetetgenele,
    readgts,
    savegts,
)
from iso2mesh.modify import (
    meshcheckrepair,
    sortmesh,
    meshresample,
    removeisolatedsurf,
    removeisolatednode,
    qmeshcut,
    removedupelem,
)

##====================================================================================
## implementations
##====================================================================================


def v2m(img, isovalues, opt=None, maxvol=None, method=None):
    """
    Volumetric mesh generation from binary or gray-scale volumetric images.

    Parameters:
    img       : 3D numpy array, volumetric image data
    isovalues : scalar or list, isovalues to generate meshes
    opt       : options for mesh generation (default: None)
    maxvol    : maximum volume for elements (default: None)
    method    : method for surface extraction, default is 'cgalsurf'

    Returns:
    node      : generated mesh nodes
    elem      : elements of the mesh
    face      : surface triangles
    """
    if method is None:
        method = "cgalsurf"

    # Generate the mesh using vol2mesh (assumes vol2mesh exists in the Python environment)
    node, elem, face = vol2mesh(
        img,
        np.arange(img.shape[0]),
        np.arange(img.shape[1]),
        np.arange(img.shape[2]),
        opt,
        maxvol,
        1,
        method,
        isovalues,
    )

    return node, elem, face


def v2s(img, isovalues, opt=None, method=None):
    """
    Surface mesh generation from binary or grayscale volumetric images.

    Parameters:
    img       : 3D numpy array, volumetric image data
    isovalues : scalar or list, isovalues to generate meshes
    opt       : options for mesh generation (default: None)
    method    : method for surface extraction, default is 'cgalsurf'

    Returns:
    no        : generated mesh nodes
    el        : elements of the mesh
    regions   : mesh regions
    holes     : mesh holes
    """
    if method is None:
        method = "cgalsurf"

    if method == "cgalmesh":
        no, tet, el = v2m(np.uint8(img), isovalues, opt, 1000, method)
        regions = []
        fclist = np.unique(el[:, 3])

        for fc in fclist:
            pt = surfinterior(no[:, :3], el[el[:, 3] == fc, :3])[0]
            if pt.size > 0:
                regions.append(pt)

        el = np.unique(el[:, :3], axis=0)
        no, el = removeisolatednode(no[:, :3], el[:, :3])
        holes = []
        return no, el, regions, holes

    no, el, regions, holes = vol2surf(
        img,
        np.arange(img.shape[0]),
        np.arange(img.shape[1]),
        np.arange(img.shape[2]),
        opt,
        1,
        method,
        isovalues,
    )

    return no, el, regions, holes


def s2m(
    v, f, keepratio=None, maxvol=None, method="tetgen", regions=None, holes=None, *args
):
    """
    Volumetric mesh generation from a closed surface, shortcut for surf2mesh.

    Parameters:
    v        : vertices of the surface
    f        : faces of the surface
    keepratio: ratio of triangles to preserve or a structure of options (for 'cgalpoly')
    maxvol   : maximum volume of mesh elements
    method   : method to use ('tetgen' by default or 'cgalpoly')
    regions  : predefined mesh regions
    holes    : holes in the mesh

    Returns:
    node     : generated mesh nodes
    elem     : elements of the mesh
    face     : surface triangles
    """
    if method == "cgalpoly":
        node, elem, face = cgals2m(v, f, keepratio, maxvol)
        return node, elem, face

    if regions is None:
        regions = []
    if holes is None:
        holes = []

    if args:
        node, elem, face = surf2mesh(
            v, f, [], [], keepratio, maxvol, regions, holes, 0, method, *args
        )
    else:
        node, elem, face = surf2mesh(
            v, f, [], [], keepratio, maxvol, regions, holes, 0, method
        )

    return node, elem, face


def s2v(node, face, div=50, **kwargs):
    """
    Convert a surface mesh to a volumetric binary image.

    Parameters:
    node   : array-like, the vertices of the triangular surface (Nx3 for x, y, z)
    face   : array-like, the triangle node indices (Mx3, each row is a triangle)
    div    : int, division number along the shortest edge of the mesh (resolution)
    kwargs  : additional arguments for the surf2vol function

    Returns:
    img    : volumetric binary image
    v2smap : 4x4 affine transformation matrix to map voxel coordinates back to the mesh space
    """
    p0 = np.min(node, axis=0)
    p1 = np.max(node, axis=0)

    if node.shape[0] == 0 or face.shape[0] == 0:
        raise ValueError("node and face cannot be empty")

    if div == 0:
        raise ValueError("div cannot be 0")

    dx = np.min(p1 - p0) / div

    if dx <= np.finfo(float).eps:
        raise ValueError("the input mesh is in a plane")

    xi = np.arange(p0[0] - dx, p1[0] + dx, dx)
    yi = np.arange(p0[1] - dx, p1[1] + dx, dx)
    zi = np.arange(p0[2] - dx, p1[2] + dx, dx)

    img = surf2vol(node, face, xi, yi, zi, **kwargs)

    return img


def vol2mesh(img, ix, iy, iz, opt, maxvol, dofix, method="cgalsurf", isovalues=None):
    """
    Convert a binary or multi-valued volume to a tetrahedral mesh.

    Parameters:
    img       : 3D numpy array, volumetric image data
    ix, iy, iz: indices for subvolume selection in x, y, z directions
    opt       : options for mesh generation
    maxvol    : maximum volume for mesh elements
    dofix     : boolean, whether to validate and repair the mesh
    method    : method for mesh generation ('cgalsurf', 'simplify', 'cgalmesh', 'cgalpoly')
    isovalues : list of isovalues for the levelset (optional)

    Returns:
    node      : node coordinates of the mesh
    elem      : element list of the mesh (last column is region ID)
    face      : surface elements of the mesh (last column is boundary ID)
    regions   : interior points for closed surfaces
    """

    if method == "cgalmesh":
        vol = img[np.ix_(ix, iy, iz)]
        if len(np.unique(vol)) > 64 and dofix == 1:
            raise ValueError(
                "CGAL mesher does not support grayscale images. Use 'cgalsurf' for grayscale volumes."
            )
        node, elem, face = cgalv2m(vol, opt, maxvol)
        return node, elem, face

    if isovalues is not None:
        no, el, regions, holes = vol2surf(
            img, ix, iy, iz, opt, dofix, method, isovalues
        )
    else:
        no, el, regions, holes = vol2surf(img, ix, iy, iz, opt, dofix, method)

    if method == "cgalpoly":
        node, elem, face = cgals2m(no[:, :3], el[:, :3], opt, maxvol)
        return node, elem, face

    node, elem, face = surf2mesh(no, el, [], [], 1, maxvol, regions, holes)
    return node, elem, face


def vol2surf(img, ix, iy, iz, opt, dofix=0, method="cgalsurf", isovalues=None):
    """
    Convert a 3D volumetric image to surfaces.

    Parameters:
    img: volumetric binary image. If img is empty, vol2surf will return user-defined surfaces via opt.surf if it exists.
    ix, iy, iz: subvolume selection indices in x, y, z directions.
    opt: options dict containing function parameters.
    dofix: 1 to perform mesh validation and repair, 0 to skip repairing.
    method: meshing method ('simplify', 'cgalsurf', or 'cgalpoly'), defaults to 'cgalsurf'.
    isovalues: list of isovalues for level sets.

    Returns:
    no: node list on the resulting surface mesh, with 3 columns for x, y, z.
    el: list of triangular elements on the surface [n1, n2, n3, region_id].
    regions: list of interior points for all sub-regions.
    holes: list of interior points for all holes.
    """

    print("Extracting surfaces from a volume...")

    if isinstance(opt, (int, float)):
        opt = {"radbound": opt}

    el = np.array([])
    no = np.array([])
    holes = opt.get("holes", [])
    regions = opt.get("regions", [])

    if img is not None and len(img) > 0:
        img = img[np.ix_(ix, iy, iz)]
        dim = img.shape
        newdim = np.array(dim) + 2
        newimg = np.zeros(newdim, dtype=img.dtype)
        newimg[1:-1, 1:-1, 1:-1] = img

        if isovalues is None:
            maxlevel = newimg.max()
            isovalues = np.arange(1, maxlevel + 1)
        else:
            tmp = np.array([isovalues]).flatten()
            isovalues = np.unique(np.sort(tmp))
            maxlevel = len(isovalues)

        for i in range(maxlevel):
            if i < maxlevel - 1:
                levelmask = (newimg >= isovalues[i]) & (newimg < isovalues[i + 1])
            else:
                levelmask = newimg >= isovalues[i]

            levelno, levelel = binsurface(levelmask, clean=False)

            if levelel.size > 0:
                if opt.get("autoregion", 0):
                    seeds = surfseeds(levelno, levelel)
                else:
                    seeds = surfinterior(levelno, levelel)[0]
                if len(seeds) > 0:
                    print(f"Region {i + 1} centroid: {seeds}")
                    regions = np.vstack((regions, seeds)) if len(regions) > 0 else seeds

        for i in range(maxlevel):
            print(f"Processing threshold level {i + 1}...")
            if method == "simplify":
                v0, f0 = binsurface(newimg >= isovalues[i])
                if dofix:
                    v0, f0 = meshcheckrepair(v0, f0)

                keepratio = (
                    opt.get("keepratio", 1)
                    if isinstance(opt, dict)
                    else opt[i].get("keepratio", 1)
                )
                print(f"Resampling surface mesh for level {i + 1}...")
                v0, f0 = meshresample(v0, f0, keepratio)
                f0 = removeisolatedsurf(v0, f0, 3)

                if dofix:
                    v0, f0 = meshcheckrepair(v0, f0)
            else:
                radbound = (
                    opt.get("radbound", 1)
                    if isinstance(opt, dict)
                    else opt[i].get("radbound", 1)
                )
                distbound = (
                    opt.get("distbound", radbound)
                    if isinstance(opt, dict)
                    else opt[i].get("distbound", radbound)
                )
                maxsurfnode = (
                    opt.get("maxnode", 40000)
                    if isinstance(opt, dict)
                    else opt[i].get("maxnode", 40000)
                )
                surfside = (
                    opt.get("side", "")
                    if isinstance(opt, dict)
                    else opt[i].get("side", "")
                )

                if surfside == "upper":
                    newimg[newimg <= isovalues[i] - 1e-9] = isovalues[i] - 1e-9
                elif surfside == "lower":
                    newimg[newimg >= isovalues[i] + 1e-9] = isovalues[i] + 1e-9

                perturb = 1e-4 * np.abs(isovalues).max()
                perturb = (
                    -perturb if np.all(newimg > isovalues[i] - perturb) else perturb
                )

                if regions.ndim == 1:
                    regions = regions[np.newaxis, :]

                v0, f0 = vol2restrictedtri(
                    newimg,
                    isovalues[i] - perturb,
                    regions[i],
                    np.sum(newdim**2) * 2,
                    30,
                    radbound,
                    distbound,
                    maxsurfnode,
                )

            if opt.get("maxsurf", 0) == 1:
                f0 = maxsurf(finddisconnsurf(f0))

            if "A" in opt and "B" in opt:
                v0 = (opt["A"] @ v0.T + opt["B"][:, None]).T

            if "hole" in opt:
                holes = np.vstack((holes, opt["hole"]))
            if "region" in opt:
                regions = np.vstack((regions, opt["region"]))

            e0 = np.hstack(
                (f0 + len(no), np.ones((f0.shape[0], 1), dtype=int) * (i + 1))
            )
            el = np.vstack((el, e0)) if len(el) > 0 else e0
            no = np.vstack((no, v0)) if len(no) > 0 else v0

    if "surf" in opt:
        for surf in opt["surf"]:
            surf["elem"][:, 3] = maxlevel + 1
            el = np.vstack((el, surf["elem"] + len(no)))
            no = np.vstack((no, surf["node"]))

    print("Surface mesh generation is complete")

    return no, el, regions, holes


# _________________________________________________________________________________________________________


def surf2mesh(
    v,
    f,
    p0,
    p1,
    keepratio,
    maxvol,
    regions=None,
    holes=None,
    dobbx=0,
    method="tetgen",
    cmdopt=None,
):
    """
    Create a quality volumetric mesh from isosurface patches.

    Parameters:
    v: isosurface node list, shape (nn,3). If v has 4 columns, the last column specifies mesh density near each node.
    f: isosurface face element list, shape (be,3). If f has 4 columns, it indicates the label of the face triangles.
    p0: coordinates of one corner of the bounding box, [x0, y0, z0].
    p1: coordinates of the other corner of the bounding box, [x1, y1, z1].
    keepratio: percentage of elements kept after simplification, between 0 and 1.
    maxvol: maximum volume of tetrahedra elements.
    regions: list of regions, specified by an internal point for each region.
    holes: list of holes, similar to regions.
    forcebox: 1 to add bounding box, 0 for automatic.
    method: meshing method (default is 'tetgen').
    cmdopt: additional options for the external mesh generator.

    Returns:
    node: node coordinates of the tetrahedral mesh.
    elem: element list of the tetrahedral mesh.
    face: mesh surface element list, with the last column denoting the boundary ID.
    """
    if keepratio > 1 or keepratio < 0:
        print(
            'The "keepratio" parameter must be between 0 and 1. No simplification will be performed.'
        )

    exesuff = getexeext()

    # Resample surface mesh if keepratio is less than 1
    if keepratio < 1 and not isinstance(f, list):
        print("Resampling surface mesh...")
        no, el = meshresample(v[:, :3], f[:, :3], keepratio)
        el = np.unique(np.sort(el, axis=1), axis=0)
    else:
        no = v
        el = f

    # Handle regions and holes arguments
    if regions is None:
        regions = np.array([])  # []
    if holes is None:
        holes = np.array([])

    # Warn if both maxvol and region-based volume constraints are specified
    if (
        isinstance(regions, np.ndarray)
        and regions.ndim > 1
        and regions.shape[1] >= 4
        and maxvol is not None
    ):
        print(
            "Warning: Both maxvol and region-based volume constraints are specified. maxvol will be ignored."
        )
        maxvol = None

    # Dump surface mesh to .poly file format
    deletemeshfile(mwpath("post_vmesh.mtr"))
    savesurfpoly(
        no, el, holes, regions, p0, p1, mwpath("post_vmesh.poly"), forcebox=dobbx
    )

    moreopt = ""
    if len(no.shape) > 1 and no.shape[1] == 4:
        moreopt = moreopt + " -m "
    # Generate volumetric mesh from surface mesh
    deletemeshfile(mwpath("post_vmesh.1.*"))
    print("Creating volumetric mesh from surface mesh...")

    if cmdopt is None:
        try:
            cmdopt = eval("ISO2MESH_TETGENOPT")
        except:
            cmdopt = ""
    cmdstr = f'"{mcpath(method, exesuff)}" -A -q1.414a{maxvol} {moreopt} ' + mwpath(
        "post_vmesh.poly"
    )
    try:
        if not sys.platform.startswith("win"):
            import tetgen

            tgen = tetgen.TetGen(filename=mwpath("post_vmesh.poly"))
            mesh = tgen.tetrahedralize(flags=cmdopt)
            node = mesh["vertices"]
            elem = mesh["tetrahedra"]
            face = mesh["faces"]
    except:
        pass

    if "mesh" not in locals():
        if not cmdopt:
            status, cmdout = subprocess.getstatusoutput(cmdstr)
        else:
            cmdstr = f'"{mcpath(method, exesuff)}" {cmdopt} ' + mwpath(
                "post_vmesh.poly"
            )
            status, cmdout = subprocess.getstatusoutput(cmdstr)

        if status != 0:
            raise RuntimeError(f"Tetgen command failed: {cmdstr}\n{cmdout}")

        # Read generated mesh
        node, elem, face = readtetgen(mwpath("post_vmesh.1"))

    print("Volume mesh generation complete")
    return node, elem, face


def surf2volz(node, face, xi, yi, zi):
    """
    Convert a triangular surface to a shell of voxels in a 3D image along the z-axis.

    Parameters:
    node: node list of the triangular surface, with 3 columns for x/y/z
    face: triangle node indices, each row represents a triangle
    xi, yi, zi: x/y/z grid for the resulting volume

    Returns:
    img: a volumetric binary image at the position of ndgrid(xi, yi, zi)
    """

    ne = face.shape[0]
    img = np.zeros((len(xi), len(yi), len(zi)), dtype=np.uint8)
    dx0 = np.min(np.abs(np.diff(xi)))
    dx = dx0 / 2
    dy0 = np.min(np.abs(np.diff(yi)))
    dy = dy0 / 2
    dz0 = np.min(np.abs(np.diff(zi)))
    dl = np.sqrt(dx**2 + dy**2)
    minz = np.min(node[:, 2])
    maxz = np.max(node[:, 2])

    # Determine the z index range
    iz, _ = np.histogram(
        [minz, maxz],
        bins=np.append(zi - np.diff(zi)[0] / 2, zi[-1] + np.diff(zi)[-1] / 2),
    )
    hz = np.nonzero(iz)[0]
    iz = np.arange(hz[0], min(len(zi), hz[-1] + 1))

    for i in iz:
        plane = np.array([[0, 100, zi[i]], [100, 0, zi[i]], [0, 0, zi[i]]])
        bcutpos, bcutvalue, bcutedges, _, _ = qmeshcut(
            face[:, :3], node, node[:, 0], plane
        )

        if bcutpos.size == 0:
            continue

        enum = bcutedges.shape[0]

        for j in range(enum):
            e0 = bcutpos[bcutedges[j, 0] - 1, :2]
            e1 = bcutpos[bcutedges[j, 1] - 1, :2]
            length = np.ceil(np.sum(np.abs(e1 - e0)) / (np.abs(dx) + np.abs(dy))) + 1
            dd = (e1 - e0) / length

            posx = np.floor(
                (e0[0] + np.arange(length + 1) * dd[0] - xi[0]) / dx0
            ).astype(int)
            posy = np.floor(
                (e0[1] + np.arange(length + 1) * dd[1] - yi[0]) / dy0
            ).astype(int)
            pos = np.vstack((posx, posy)).T

            pos = pos[(posx > 0) & (posx <= len(xi)) & (posy > 0) & (posy <= len(yi))]

            if len(pos) > 0:
                zz = np.floor((zi[i] - zi[0]) / dz0).astype(int)
                for k in range(pos.shape[0]):
                    img[pos[k, 0], pos[k, 1], zz] = 1

    return img


def surf2vol(node, face, xi, yi, zi, **kwargs):
    """
    Convert a triangular surface to a shell of voxels in a 3D image.

    Parameters:
    node: node list of the triangular surface, 3 columns for x/y/z
    face: triangle node indices, each row is a triangle
          If face contains a 4th column, it indicates the label of the face triangles.
          If face contains 5 columns, it stores a tetrahedral mesh with labels.
    xi, yi, zi: x/y/z grid for the resulting volume
    kwargs: optional parameters:
        'fill': if set to 1, the enclosed voxels are labeled as 1.
        'label': if set to 1, the enclosed voxels are labeled by the corresponding label of the face or element.
                 Setting 'label' to 1 also implies 'fill'.

    Returns:
    img: a volumetric binary image at the position of ndgrid(xi, yi, zi)
    v2smap (optional): a 4x4 matrix denoting the Affine transformation to map voxel coordinates back to the mesh space.
    """
    from scipy.ndimage import binary_fill_holes

    opt = kwargs
    label = opt.get("label", 0)
    elabel = np.array([1], dtype=np.int32)

    if face.shape[1] >= 4:
        elabel = np.unique(face[:, -1])
        if face.shape[1] == 5:
            label = 1
            el = face
            face = np.empty((0, 4))
            for lbl in elabel:
                fc = volface(el[el[:, 4] == lbl, :4])
                fc = np.hstack((fc, np.full((fc.shape[0], 1), lbl)))
                face = np.vstack((face, fc))
    else:
        fc = face

    img = np.zeros((len(xi), len(yi), len(zi)), dtype=elabel.dtype)

    for lbl in elabel:
        if face.shape[1] == 4:
            fc = face[face[:, 3] == lbl, :3]

        im = surf2volz(node[:, :3], fc[:, :3], xi, yi, zi)
        im |= np.moveaxis(surf2volz(node[:, [2, 0, 1]], fc[:, :3], zi, xi, yi), 0, 2)
        im |= np.moveaxis(
            surf2volz(node[:, [1, 2, 0]], fc[:, :3], yi, zi, xi), [0, 1], [-2, -1]
        )

        if opt.get("fill", 0) or label:
            im = binary_fill_holes(im)
            if label:
                im = im.astype(elabel.dtype) * lbl

        img = np.maximum(im.astype(img.dtype), img)

    if "v2smap" in kwargs:
        dlen = np.abs([xi[1] - xi[0], yi[1] - yi[0], zi[1] - zi[0]])
        offset = np.min(node, axis=0)
        v2smap = np.eye(4)
        v2smap[:3, :3] = np.diag(np.abs(dlen))
        v2smap[:3, 3] = offset
        return img.astype(np.uint8), v2smap

    return img.astype(np.uint8)


def binsurface(img, nface=3, **kwargs):
    """
    node, elem = binsurface(img, nface)

    Fast isosurface extraction from 3D binary images.

    Parameters:
        img: 3D binary NumPy array
        nface:
            = 3 or omitted: triangular faces
            = 4: square (quad) faces
            = 0: return boundary mask image via `node`
            = 'iso': use marching cubes (`isosurface` equivalent)

    Returns:
        node: (N, 3) array of vertex coordinates
        elem: (M, 3) or (M, 4) array of face elements (1-based indices)
    """
    if isinstance(nface, str) and nface == "iso":
        from skimage.measure import marching_cubes

        verts, faces, _, _ = marching_cubes(img, level=0.5)
        node = verts[:, [1, 0, 2]] - 0.5  # reorder to match MATLAB
        elem = faces + 1  # 1-based indexing
        return node, elem

    dim = list(img.shape)
    if len(dim) < 3:
        dim += [1]
    newdim = [d + 1 for d in dim]

    # Compute differences in each direction
    d1 = np.diff(img, axis=0)
    d2 = np.diff(img, axis=1)
    d3 = np.diff(img, axis=2)

    ix, iy, iz = np.where((d1 == 1) | (d1 == -1))
    jx, jy, jz = np.where((d2 == 1) | (d2 == -1))
    kx, ky, kz = np.where((d3 == 1) | (d3 == -1))

    ix += 1
    jy += 1
    kz += 1

    # Adjust indices and wrap them to 3D
    id1 = np.ravel_multi_index((ix, iy, iz), newdim, order="F")
    id2 = np.ravel_multi_index((jx, jy, jz), newdim, order="F")
    id3 = np.ravel_multi_index((kx, ky, kz), newdim, order="F")

    if nface == 0:
        elem = np.concatenate([id1, id2, id3])
        node = np.zeros(newdim, dtype=np.int8)
        node.flat[elem] = 1
        node = node[1:-1, 1:-1, 1:-1] - 1
        return node, elem

    xy = newdim[0] * newdim[1]

    if nface == 3:  # triangles
        elem = np.vstack(
            [
                np.column_stack([id1, id1 + newdim[0], id1 + newdim[0] + xy]),
                np.column_stack([id1, id1 + newdim[0] + xy, id1 + xy]),
                np.column_stack([id2, id2 + 1, id2 + 1 + xy]),
                np.column_stack([id2, id2 + 1 + xy, id2 + xy]),
                np.column_stack([id3, id3 + 1, id3 + 1 + newdim[0]]),
                np.column_stack([id3, id3 + 1 + newdim[0], id3 + newdim[0]]),
            ]
        )
    else:  # quads
        elem = np.vstack(
            [
                np.column_stack([id1, id1 + newdim[0], id1 + newdim[0] + xy, id1 + xy]),
                np.column_stack([id2, id2 + 1, id2 + 1 + xy, id2 + xy]),
                np.column_stack([id3, id3 + 1, id3 + 1 + newdim[0], id3 + newdim[0]]),
            ]
        )

    # Compress the node indices
    maxid = np.max(elem) + 1
    nodemap = np.zeros(maxid, dtype=int)
    nodemap[elem.ravel(order="F")] = 1
    id = np.where(nodemap)[0]

    # Reindex elem to be compact and 1-based
    nodemap = np.zeros_like(nodemap)
    nodemap[id] = np.arange(1, len(id) + 1)  # 1-based
    elem = nodemap[elem]

    # Create node coordinates
    xi, yi, zi = np.unravel_index(id, newdim, order="F")
    node = np.column_stack([xi, yi, zi])

    if nface == 3 and kwargs.get("clean", True):
        node, elem = meshcheckrepair(node, elem)

    return node, elem


def cgalv2m(vol, opt, maxvol):
    """
    Wrapper for CGAL 3D mesher (CGAL 3.5 or up) to convert a binary or multi-valued volume to tetrahedral mesh.

    Parameters:
    vol: a volumetric binary image.
    opt: parameters for the CGAL mesher. If opt is a structure:
        opt.radbound: maximum surface element size.
        opt.angbound: minimum angle of a surface triangle.
        opt.distbound: maximum distance between the center of the surface bounding circle and the element bounding sphere.
        opt.reratio: maximum radius-edge ratio.
        If opt is a scalar, it specifies radbound.
    maxvol: target maximum tetrahedral element volume.

    Returns:
    node: node coordinates of the tetrahedral mesh.
    elem: element list of the tetrahedral mesh, the last column is the region ID.
    face: mesh surface element list of the tetrahedral mesh, the last column denotes the boundary ID.
    """

    print("Creating surface and tetrahedral mesh from a multi-domain volume...")

    if not (np.issubdtype(vol.dtype, np.bool_) or vol.dtype == np.uint8):
        raise ValueError(
            "CGAL mesher can only handle uint8 volumes. Convert your image to uint8 first."
        )

    if not np.any(vol):
        raise ValueError("No labeled regions found in the input volume.")

    exesuff = getexeext()
    exesuff = fallbackexeext(exesuff, "cgalmesh")

    ang = 30
    ssize = 6
    approx = 0.5
    reratio = 3

    if not isinstance(opt, dict):
        ssize = opt

    if isinstance(opt, dict):
        ssize = opt.get("radbound", ssize)
        ang = opt.get("angbound", ang)
        approx = opt.get("distbound", approx)
        reratio = opt.get("reratio", reratio)

    saveinr(vol, mwpath("pre_cgalmesh.inr"))
    deletemeshfile(mwpath("post_cgalmesh.mesh"))

    randseed = int("623F9A9E", 16)

    cmd = f'"{mcpath("cgalmesh", exesuff)}" "{mwpath("pre_cgalmesh.inr")}" "{mwpath("post_cgalmesh.mesh")}" {ang} {ssize} {approx} {reratio} {maxvol} {randseed}'

    status, str_output = subprocess.getstatusoutput(cmd)

    if not os.path.exists(mwpath("post_cgalmesh.mesh")):
        raise RuntimeError(f"Output file was not found. Command failed: {cmd}")

    node, elem, face = readmedit(mwpath("post_cgalmesh.mesh"))

    if isinstance(opt, dict) and len(opt) == 1:
        if "A" in opt and "B" in opt:
            node[:, :3] = (
                opt["A"] @ node[:, :3].T
                + np.tile(opt["B"][:, None], (1, node.shape[0]))
            ).T

    print(
        f"Node number: {node.shape[0]}\nTriangles: {face.shape[0]}\nTetrahedra: {elem.shape[0]}\nRegions: {len(np.unique(elem[:, -1]))}"
    )
    print("Surface and volume meshes complete")

    if node.shape[0] > 0:
        node, elem, face, _ = sortmesh(
            node[0, :], node, elem, list(range(4)), face, list(range(3))
        )

    node += 0.5
    elem[:, :4] = meshreorient(node[:, :3], elem[:, :4])[0]

    return node, elem, face


def cgals2m(v, f, opt, maxvol, **kwargs):
    """
    Convert a triangular surface to a tetrahedral mesh using CGAL mesher.

    Parameters:
    v : ndarray
        Node coordinate list of a surface mesh (nn x 3)
    f : ndarray
        Face element list of a surface mesh (be x 3)
    opt : dict or scalar
        Parameters for CGAL mesher. If it's a dict, it can include:
            - radbound: Maximum surface element size
            - angbound: Minimum angle of a surface triangle
            - distbound: Max distance between surface bounding circle center and element bounding sphere center
            - reratio: Maximum radius-edge ratio
        If it's a scalar, it only specifies radbound.
    maxvol : float
        Target maximum tetrahedral element volume.
    kwargs : Additional arguments

    Returns:
    node : ndarray
        Node coordinates of the tetrahedral mesh.
    elem : ndarray
        Element list of the tetrahedral mesh. The last column is the region id.
    face : ndarray
        Mesh surface element list of the tetrahedral mesh. The last column denotes the boundary ID.
    """

    print("Creating surface and tetrahedral mesh from a polyhedral surface ...")

    exesuff = fallbackexeext(getexeext(), "cgalpoly")

    ang = 30
    ssize = 6
    approx = 0.5
    reratio = 3

    if not isinstance(opt, dict):
        ssize = opt

    if isinstance(opt, dict):
        ssize = opt.get("radbound", ssize)
        ang = opt.get("angbound", ang)
        approx = opt.get("distbound", approx)
        reratio = opt.get("reratio", reratio)

    if kwargs.get("dorepair", 0) == 1:
        v, f = meshcheckrepair(v, f)

    saveoff(v, f, mwpath("pre_cgalpoly.off"))
    deletemeshfile(mwpath("post_cgalpoly.mesh"))

    randseed = os.getenv("ISO2MESH_SESSION", int("623F9A9E", 16))

    cmd = (
        f'"{mcpath("cgalpoly", exesuff)}" "{mwpath("pre_cgalpoly.off")}" "{mwpath("post_cgalpoly.mesh")}" '
        f"{ang:.16f} {ssize:.16f} {approx:.16f} {reratio:.16f} {maxvol:.16f} {randseed}"
    )

    status = subprocess.call(cmd, shell=True)

    if status != 0:
        raise RuntimeError("cgalpoly command failed")

    if not os.path.exists(mwpath("post_cgalpoly.mesh")):
        raise FileNotFoundError(
            f"Output file was not found, failure occurred when running command: \n{cmd}"
        )

    node, elem, face = readmedit(mwpath("post_cgalpoly.mesh"))

    print(f"node number:\t{node.shape[0]}")
    print(f"triangles:\t{face.shape[0]}")
    print(f"tetrahedra:\t{elem.shape[0]}")
    print(f"regions:\t{len(np.unique(elem[:, -1]))}")
    print("Surface and volume meshes complete")

    return node, elem, face


def getintersecttri(tmppath):
    """
    Get the IDs of self-intersecting elements from TetGen.

    Args:
    tmppath: Working directory where TetGen output is stored.

    Returns:
    eid: An array of all intersecting surface element IDs.
    """
    exesuff = getexeext()
    exesuff = fallbackexeext(exesuff, "tetgen")
    tetgen_path = mcpath("tetgen", exesuff)

    command = f'"{tetgen_path}" -d "{os.path.join(tmppath, "post_vmesh.poly")}"'
    status, str_output = subprocess.getstatusoutput(command)

    eid = []
    if status == 0:
        ids = re.findall(r" #([0-9]+) ", str_output)
        eid = [int(id[0]) for id in ids]

    eid = np.unique(eid)
    return eid


def vol2restrictedtri(vol, thres, cent, brad, ang, radbound, distbound, maxnode):
    """
    Surface mesh extraction using CGAL mesher.

    Parameters:
    vol : ndarray
        3D volumetric image.
    thres : float
        Threshold for extraction.
    cent : tuple
        A 3D position (x, y, z) inside the resulting mesh.
    brad : float
        Maximum bounding sphere squared of the resulting mesh.
    ang : float
        Minimum angular constraint for triangular elements (degrees).
    radbound : float
        Maximum triangle Delaunay circle radius.
    distbound : float
        Maximum Delaunay sphere distances.
    maxnode : int
        Maximum number of surface nodes.

    Returns:
    node : ndarray
        List of 3D nodes (x, y, z) in the resulting surface.
    elem : ndarray
        Element list of the resulting mesh (3 columns of integers).
    """

    if radbound < 1:
        print(
            "You are meshing the surface with sub-pixel size. Check if opt.radbound is set correctly."
        )

    exesuff = getexeext()

    # Save the input volume in .inr format
    saveinr(vol, mwpath("pre_extract.inr"))

    # Delete previous output mesh file if exists
    deletemeshfile(mwpath("post_extract.off"))

    # Random seed
    randseed = os.getenv("ISO2MESH_SESSION", int("623F9A9E", 16))

    initnum = os.getenv("ISO2MESH_INITSIZE", 50)

    # Build the system command to run CGAL mesher
    cmd = (
        f'"{mcpath("cgalsurf", exesuff)}" "{mwpath("pre_extract.inr")}" '
        f"{thres:.16f} {cent[0]:.16f} {cent[1]:.16f} {cent[2]:.16f} {brad:.16f} {ang:.16f} {radbound:.16f} "
        f'{distbound:.16f} {maxnode} "{mwpath("post_extract.off")}" {randseed} {initnum}'
    )

    # Execute the system command
    status = subprocess.call(cmd, shell=True)
    if status != 0:
        raise RuntimeError(f"CGAL mesher failed with command: {cmd}")

    # Read the resulting mesh
    node, elem = readoff(mwpath("post_extract.off"))
    # Check and repair mesh if needed
    node, elem = meshcheckrepair(node, elem)

    # Assuming the origin [0, 0, 0] is located at the lower-bottom corner of the image
    node += 0.5

    return node, elem


def fillsurf(node, face):
    """
    Calculate the enclosed volume for a closed surface mesh.

    Args:
        node: Node coordinates (nn, 3).
        face: Surface triangle list (ne, 3).

    Returns:
        no: Node coordinates of the filled volume mesh.
        el: Element list (tetrahedral elements) of the filled volume mesh.
    """

    # Placeholder for calling an external function, typically using TetGen for surface to volume mesh conversion
    no, el, _ = surf2mesh(node, face, None, None, 1, 1, None, None, 0, "tetgen", "-YY")

    return no, el


def outersurf(node, face):
    """
    Extract the outer-most shell of a complex surface mesh.

    Parameters:
    node: Node coordinates
    face: Surface triangle list

    Returns:
    outface: The outer-most shell of the surface mesh
    """

    face = face[:, :3]  # Limit face to first 3 columns
    ed = surfedge(face)[0]  # Find surface edges

    # If surface is open, raise an error
    if ed.size != 0:
        raise ValueError(
            "Open surface detected, close it first. Consider meshcheckrepair() with meshfix option."
        )

    # Fill the surface and extract the volume's outer surface
    no, el = fillsurf(node, face)
    outface = volface(el)

    # Remove isolated nodes
    no, outface = removeisolatednode(no, outface)

    # Check matching of node coordinates
    maxfacenode = np.max(outface)
    I, J = ismember_rows(np.round(no[:maxfacenode, :] * 1e10), np.round(node * 1e10))

    # Map faces to the original node set
    outface = J[outface]

    # Remove faces with unmapped (zero-indexed) nodes
    ii, jj = np.where(outface == 0)
    outface = np.delete(outface, ii, axis=0)

    return outface


def surfvolume(node, face, option=None):
    """
    Calculate the enclosed volume for a closed surface.

    Parameters:
    node: Node coordinates
    face: Surface triangle list
    option: (Optional) additional option, currently unused

    Returns:
    vol: Total volume of the enclosed space
    """

    face = face[:, :3]  # Limit face to first 3 columns
    ed = surfedge(face)[0]  # Detect surface edges

    # If surface is open, raise an error
    if ed.size != 0:
        raise ValueError(
            "Open surface detected, you need to close it first. Consider meshcheckrepair() with the meshfix option."
        )

    # Fill the surface and calculate the volume of enclosed elements
    no, el = fillsurf(node, face)
    vol = elemvolume(no, el)

    # Sum the volume of all elements
    vol = np.sum(vol)

    return vol


def insurface(node, face, points):
    """
    Test if a set of 3D points is located inside a 3D triangular surface.

    Parameters:
    node: Node coordinates (Nx3 array)
    face: Surface triangle list (Mx3 array)
    points: A set of 3D points to test (Px3 array)

    Returns:
    tf: A binary vector of length equal to the number of points.
        1 indicates the point is inside the surface, and 0 indicates outside.
    """

    from scipy.spatial import Delaunay

    # Fill the surface and get nodes and elements
    no, el = fillsurf(node, face)

    # Check if points are inside the surface using Delaunay triangulation
    tri = Delaunay(no)
    tf = tri.find_simplex(points) >= 0

    # Set points inside the surface to 1, and outside to 0
    tf = tf.astype(int)

    return tf


def remeshsurf(node, face, opt):
    """
    remeshsurf(node, face, opt)

    Remesh a triangular surface, output is guaranteed to be free of self-intersecting elements.
    This function can both downsample or upsample a mesh.

    Parameters:
        node: list of nodes on the input surface mesh, 3 columns for x, y, z
        face: list of triangular elements on the surface, [n1, n2, n3, region_id]
        opt: function parameters
            opt.gridsize: resolution for the voxelization of the mesh
            opt.closesize: if there are openings, set the closing diameter
            opt.elemsize: the size of the element of the output surface
            If opt is a scalar, it defines the elemsize and gridsize = opt / 4

    Returns:
        newno: list of nodes on the resulting surface mesh, 3 columns for x, y, z
        newfc: list of triangular elements on the surface, [n1, n2, n3, region_id]
    """
    from scipy.ndimage import binary_fill_holes

    # Step 1: convert the old surface to a volumetric image
    p0 = np.min(node, axis=0)
    p1 = np.max(node, axis=0)

    if isinstance(opt, dict):
        dx = opt.get("gridsize", None)
    else:
        dx = opt / 4

    x_range = np.arange(p0[0] - dx, p1[0] + dx, dx)
    y_range = np.arange(p0[1] - dx, p1[1] + dx, dx)
    z_range = np.arange(p0[2] - dx, p1[2] + dx, dx)

    img = surf2vol(node, face, x_range, y_range, z_range)

    # Compute surface edges
    eg = surfedge(face)[0]

    closesize = 0
    if eg.size > 0 and isinstance(opt, dict):
        closesize = opt.get("closesize", 0)

    # Step 2: fill holes in the volumetric binary image
    img = binary_fill_holes(img)

    # Step 3: convert the filled volume to a new surface
    if isinstance(opt, dict):
        if "elemsize" in opt:
            opt["radbound"] = opt["elemsize"] / dx
            newno, newfc, _, _ = v2s(img, 0.5, opt, "cgalsurf")
    else:
        opt = {"radbound": opt / dx}
        newno, newfc, _, _ = v2s(img, 0.5, opt, "cgalsurf")

    # Adjust new nodes to match original coordinates
    newno[:, 0:3] *= dx
    newno[:, 0] += p0[0]
    newno[:, 1] += p0[1]
    newno[:, 2] += p0[2]

    return newno, newfc


def meshrefine(node, elem, *args, **kwargs):
    """
    meshrefine - refine a tetrahedral mesh by adding new nodes or constraints

    Usage:
        newnode, newelem, newface = meshrefine(node, elem, face=None, opt=None)

    Description:
        This function refines a tetrahedral mesh by inserting new nodes or applying
        sizing constraints.

    Inputs:
        node: Nx3 or Nx4 array of mesh nodes. If Nx4, the 4th column is used as a
              size field.
        elem: Mx4 or Mx5 array of tetrahedral elements (1-based indexing).
        face (optional): Px3 array of triangle faces (1-based indexing).
        opt: options for mesh refinement. Can be one of:
             - Nx3 array of new nodes to be inserted (must be inside or on the mesh)
             - Vector of desired edge-lengths (length = len(node)) or max volumes (len = len(elem))
             - Dictionary with fields:
                - newnode: same as above Nx3 array
                - reratio: radius-edge ratio (default 1.414)
                - maxvol: max tetrahedron volume
                - sizefield: node-based or element-based sizing
                - extcmdopt: tetgen options for inserting external nodes
                - extlabel: label for new external elements (default: 0)
                - extcorelabel: label for core of external mesh (default: -1)

    Outputs:
        newnode: refined node list
        newelem: refined element list
        newface: surface faces (Px3), last column denotes boundary ID

    Examples:
        # Inserting internal nodes
        innernodes = np.array([[1,1,1], [2,2,2], [3,3,3]])
        newnode, newelem, _ = meshrefine(node, elem, innernodes)

        # Inserting external nodes
        extnodes = np.array([[-5,-5,25], [-5,5,25], [5,5,25], [5,-5,25]])
        opt = {'newnode': extnodes, 'extcmdopt': '-Y'}
        newnode, newelem, _ = meshrefine(node, elem, opt)

    Author:
        Qianqian Fang <q.fang@neu.edu>

    License:
        Part of Iso2Mesh Toolbox (http://iso2mesh.sf.net)
    """
    newpt = None
    sizefield = None
    face = None
    opt = {}

    if node.shape[1] == 4:
        sizefield = node[:, 3]
        node = node[:, :3]

    if len(args) == 1:
        if isinstance(args[0], dict):
            opt = args[0]
        elif len(args[0]) == node.shape[0] or len(args[0]) == elem.shape[0]:
            sizefield = np.array(args[0])
        else:
            newpt = np.array(args[0])
    elif len(args) >= 2:
        face = np.array(args[0])
        if isinstance(args[1], dict):
            opt = args[1]
        elif len(args[1]) == node.shape[0] or len(args[1]) == elem.shape[0]:
            sizefield = np.array(args[1])
        else:
            newpt = np.array(args[1])

    if newpt is None and "newnode" in kwargs:
        newpt = np.array(kwargs["newnode"])
    if sizefield is None and "sizefield" in kwargs:
        sizefield = np.array(kwargs["sizefield"])

    exesuff = fallbackexeext(getexeext(), "tetgen")

    deletemeshfile(mwpath("pre_refine.*"))
    deletemeshfile(mwpath("post_refine.*"))

    moreopt = ""
    setquality = False

    if "reratio" in kwargs:
        moreopt += f" -q {kwargs['reratio']:.10f} "
        setquality = True
    if "maxvol" in kwargs:
        moreopt += f" -a{kwargs['maxvol']:.10f} "

    externalpt = np.empty((0, 3))
    if "extcmdopt" in kwargs and newpt is not None:
        from scipy.spatial import Delaunay

        try:
            tri = Delaunay(node[:, :3])
            isinside = tri.find_simplex(newpt[:, :3]) >= 0
        except Exception:
            isinside = np.zeros(newpt.shape[0], dtype=bool)
        externalpt = newpt[~isinside]
        newpt = newpt[isinside]

    if newpt is not None and len(newpt) > 0 and newpt.shape[0] < 4:
        newpt = np.vstack([newpt, np.tile(newpt[0], (4 - newpt.shape[0], 1))])

    if newpt is not None and len(newpt) > 0:
        savetetgennode(newpt, mwpath("pre_refine.1.a.node"))
        moreopt = " -i "

    if sizefield is not None:
        if len(sizefield) == node.shape[0]:
            with open(mwpath("pre_refine.1.mtr"), "w") as f:
                f.write(f"{len(sizefield)} 1\n")
                np.savetxt(f, sizefield, fmt="%.16g")
            moreopt += " -qma "
        else:
            with open(mwpath("pre_refine.1.vol"), "w") as f:
                f.write(f"{len(sizefield)}\n")
                rownum = np.arange(1, sizefield.size + 1).reshape(-1, 1)
                np.savetxt(
                    f, np.hstack((rownum, sizefield.reshape(-1, 1)), fmt="%d %.16g")
                )
            moreopt += " -qma "

    if elem.shape[1] == 3 and not setquality:
        if newpt is not None:
            raise ValueError("Inserting new point cannot be used for surfaces")
        nedge = savegts(node, elem, mwpath("pre_refine.gts"))
        exesuff = fallbackexeext(getexeext(), "gtsrefine")
    elif elem.shape[1] == 3:
        savesurfpoly(node, elem, None, None, None, None, mwpath("pre_refine.poly"))
    else:
        savetetgennode(node, mwpath("pre_refine.1.node"))
        savetetgenele(elem, mwpath("pre_refine.1.ele"))

    print("refining the input mesh ...")

    if elem.shape[1] == 3 and not setquality:
        if isinstance(opt, dict) and "scale" in opt:
            moreopt += f" -n {int(round(nedge * opt['scale']))} "
        else:
            raise ValueError("You must give opt.scale value for refining a surface")

    if isinstance(opt, dict) and "moreopt" in opt:
        moreopt += opt["moreopt"]

    if elem.shape[1] == 3 and not setquality:
        status, cmdout = subprocess.getstatusoutput(
            f'"{mcpath("gtsrefine", exesuff)}" {moreopt} < "{mwpath("pre_refine.gts")}" > "{mwpath("post_refine.gts")}"'
        )
        if status:
            raise RuntimeError("gtsrefine command failed")
        newnode, newelem = readgts(mwpath("post_refine.gts"))
        newface = newelem
    elif elem.shape[1] == 3:
        status, cmdout = subprocess.getstatusoutput(
            f'"{mcpath("tetgen1.5", exesuff)}" {moreopt} -p -A "{mwpath("pre_refine.poly")}"'
        )
        if status:
            raise RuntimeError("tetgen command failed")
        newnode, newelem, newface = readtetgen(mwpath("pre_refine.1"))
    elif moreopt:
        status, cmdout = subprocess.getstatusoutput(
            f'"{mcpath("tetgen", exesuff)}" {moreopt} -r "{mwpath("pre_refine.1")}"'
        )
        if status:
            raise RuntimeError("tetgen command failed")
        newnode, newelem, newface = readtetgen(mwpath("pre_refine.2"))
    else:
        newnode = node
        newelem = elem
        newface = face

    if (
        externalpt.size > 0
    ):  # user request to insert nodes that are outside of the original mesh
        from scipy.spatial import ConvexHull, Delaunay

        # create a mesh including the external points
        externalpt = np.unique(externalpt, axis=0)
        allnode = np.vstack([newnode, externalpt])

        # define the convex hull as the external surface
        outface = ConvexHull(allnode, qhull_options="QJ").simplices + 1
        outface = np.sort(outface, axis=1)

        face = volface(newelem[:, :4])[0]  # adjust for 1-based indexing
        inface = np.sort(face[:, :3], axis=1)

        # define the surface that bounds the newly extended convex hull space
        bothsides = removedupelem(np.vstack([outface, inface]))

        # define a seed point to avoid meshing the interior space
        holelist = surfseeds(newnode, face[:, :3])

        # mesh the extended space
        ISO2MESH_TETGENOPT = kwargs.get("extcmdopt", "-Y")
        try:
            if bothsides.shape[0] >= inface.shape[0]:
                no, el, _ = surf2mesh(
                    allnode,
                    bothsides,
                    None,
                    None,
                    1,
                    10,
                    None,
                    holelist,
                    0,
                    "tetgen",
                    ISO2MESH_TETGENOPT,
                )
            else:
                no, el, _ = surf2mesh(
                    allnode,
                    bothsides,
                    None,
                    None,
                    1,
                    10,
                    None,
                    None,
                    0,
                    "tetgen",
                    ISO2MESH_TETGENOPT,
                )
        except:
            bothsides = maxsurf(finddisconnsurf(bothsides), allnode)[0]
            if bothsides.shape[0] >= inface.shape[0]:
                no, el, _ = surf2mesh(
                    allnode,
                    bothsides,
                    None,
                    None,
                    1,
                    10,
                    None,
                    holelist,
                    0,
                    "tetgen",
                    ISO2MESH_TETGENOPT,
                )
            else:
                no, el, _ = surf2mesh(
                    allnode,
                    bothsides,
                    None,
                    None,
                    1,
                    10,
                    None,
                    None,
                    0,
                    "tetgen",
                    ISO2MESH_TETGENOPT,
                )

        # map the new node coordinates back to the original node list
        rounded_no = np.round(no, 10)
        rounded_allnode = np.round(allnode, 10)
        _, map = ismember_rows(rounded_no, rounded_allnode)
        snid = np.arange(len(newnode) + 1, len(allnode) + 1)  # 0-based indexing

        # add unmapped nodes
        if np.any(map == 0):
            missing = no[map == 0]
            oldsize = allnode.shape[0]
            allnode = np.vstack([allnode, missing])
            map[map == 0] = np.arange(oldsize, allnode.shape[0]) + 1

        # merge the external space with the original mesh
        el2 = map[el[:, :4] - 1]

        # label all new elements with -1
        if newelem.shape[1] == 5:
            extlabel = jsonopt("extlabel", 0, opt)
            extcorelabel = jsonopt("extcorelabel", -1, opt)
            fifth_col = np.full((el2.shape[0], 1), extlabel)
            el2 = np.hstack([el2, fifth_col])
            iselm = np.isin(el2[:, :4], snid).astype(int)
            el2[np.sum(iselm, axis=1) >= 3, 4] = extcorelabel

        # merge nodes/elements and replace the original ones
        newnode = allnode
        newelem = np.vstack([newelem, el2])

    print("mesh refinement is complete")
    return newnode, newelem, newface
